from collections.abc import Callable
from pathlib import Path

import polars as pl

from .types import DF_T

WRITE_FN_T = Callable[[DF_T, Path], None]


def write_df(df: DF_T, out_fp: Path) -> None:
    """A generic helper to write a dataframe, either lazy or eager, to a parquet file."""
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    out_fp.parent.mkdir(parents=True, exist_ok=True)
    df.write_parquet(out_fp, use_pyarrow=True)



"""Locking functions."""

import logging
from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path

import pyarrow.parquet as pq
from filelock import FileLock, Timeout

from ..compute_modes import COMPUTE_FN_T
from ..dataframe import READ_FN_T, WRITE_FN_T

logger = logging.getLogger(__name__)

LOCK_TIME_FMT = "%Y-%m-%dT%H:%M:%S.%f"
FILE_CHECKER_T = Callable[[Path], bool]


def is_complete_parquet_file(fp: Path) -> bool:
    """Check if a parquet file is complete.

    Args:
        fp: The file path to the parquet file.

    Returns:
        True if the parquet file is complete, False otherwise.

    Examples:
        >>> df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> with tempfile.NamedTemporaryFile() as tmp:
        ...     df.write_parquet(tmp)
        ...     is_complete_parquet_file(tmp)
        True
        >>> with tempfile.NamedTemporaryFile() as tmp:
        ...     df.write_csv(tmp)
        ...     is_complete_parquet_file(tmp)
        False
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     tmp = Path(tmp)
        ...     is_complete_parquet_file(tmp / "nonexistent.parquet")
        False
    """

    try:
        _ = pq.ParquetFile(fp)
        return True
    except Exception:
        return False


def default_file_checker(fp: Path) -> bool:
    """Check if a file exists and is complete."""
    if fp.suffix == ".parquet":
        return is_complete_parquet_file(fp)
    return fp.is_file()


def rwlock_wrap(
    in_fp: Path,
    out_fp: Path,
    read_fn: READ_FN_T,
    write_fn: WRITE_FN_T,
    compute_fn: COMPUTE_FN_T,
    do_overwrite: bool = False,
    out_fp_checker: FILE_CHECKER_T = default_file_checker,
) -> bool:
    """Wrap a series of file-in file-out map transformations on a dataframe with caching and locking.

    Args:
        in_fp: The file path of the input dataframe. Must exist and be readable via `read_fn`.
        out_fp: Output file path. The parent directory will be created if it does not exist. If this file
            already exists, it will be deleted before any computations are done if `do_overwrite=True`, which
            can result in data loss if the transformation functions do not complete successfully on
            intermediate steps. If `do_overwrite` is `False` and this file exists, the function will use the
            `read_fn` to read the file and return the dataframe directly.
        read_fn: Function that reads the dataframe from a file. This must take as input a Path object and
            return a dataframe. Ideally, this read function can make use of lazy
            loading to further accelerate unnecessary reads when resuming from intermediate cached steps.
        write_fn: Function that writes the dataframe to a file. This must take as input a dataframe and a Path
            object, and will write the dataframe to that file.
        compute_fn: A function that transform the dataframe, which must take as input and return a dataframe.
        do_overwrite: If True, the output file will be overwritten if it already exists. This is `False` by
            default.

    Returns:
        True if the computation was run, False otherwise.

    Examples:
        >>> directory = tempfile.TemporaryDirectory()
        >>> read_fn = pl.read_csv
        >>> write_fn = pl.DataFrame.write_csv
        >>> root = Path(directory.name)
        >>> # For this example we'll use a simple CSV file, but in practice we *strongly* recommend using
        >>> # Parquet files for performance reasons.
        >>> in_fp = root / "input.csv"
        >>> out_fp = root / "output.csv"
        >>> in_df = pl.DataFrame({"a": [1, 3, 3], "b": [2, 4, 5], "c": [3, -1, 6]})
        >>> in_df.write_csv(in_fp)
        >>> def compute_fn(df: pl.DataFrame) -> pl.DataFrame:
        ...     return df.with_columns(pl.col("c") * 2).filter(pl.col("c") > 4)
        >>> result_computed = rwlock_wrap(in_fp, out_fp, read_fn, write_fn, compute_fn)
        >>> assert result_computed
        >>> print(out_fp.read_text())
        a,b,c
        1,2,6
        3,5,12
        <BLANKLINE>
        >>> in_df_2 = pl.DataFrame({"a": [1], "b": [3], "c": [-1]})
        >>> in_fp_2 = root / "input_2.csv"
        >>> in_df_2.write_csv(in_fp_2)
        >>> compute_fn = lambda df: df
        >>> result_computed = rwlock_wrap(in_fp_2, out_fp, read_fn, write_fn, compute_fn, do_overwrite=True)
        >>> assert result_computed
        >>> print(out_fp.read_text())
        a,b,c
        1,3,-1
        <BLANKLINE>
        >>> out_fp.unlink()
        >>> compute_fn = lambda df: df.with_columns(pl.col("c") * 2).filter(pl.col("d") > 4)
        >>> rwlock_wrap(in_fp, out_fp, read_fn, write_fn, compute_fn)
        Traceback (most recent call last):
            ...
        polars.exceptions.ColumnNotFoundError: unable to find column "d"; valid columns: ["a", "b", "c"]
        >>> assert not out_fp.is_file()  # Out file should not be created when the process crashes

    If the lock file already exists, the function will not do anything

        >>> def compute_fn(df: pl.DataFrame) -> pl.DataFrame:
        ...     return df.with_columns(pl.col("c") * 2).filter(pl.col("c") > 4)
        >>> out_fp = root / "output.csv"
        >>> lock_fp = root / "output.csv.lock"
        >>> with FileLock(str(lock_fp)):
        ...     result_computed = rwlock_wrap(in_fp, out_fp, read_fn, write_fn, compute_fn)
        ...     assert not result_computed

    The lock file will be removed after successful processing.

        >>> result_computed = rwlock_wrap(in_fp, out_fp, read_fn, write_fn, compute_fn)
        >>> assert result_computed
        >>> assert not lock_fp.exists()
    """

    if out_fp_checker(out_fp):
        if do_overwrite:
            logger.info(f"Deleting existing {out_fp} as do_overwrite={do_overwrite}.")
            out_fp.unlink()
        else:
            logger.info(f"{out_fp} exists; returning.")
            return False

    lock_fp = out_fp.with_suffix(f"{out_fp.suffix}.lock")
    lock = FileLock(str(lock_fp))
    try:
        lock.acquire(timeout=0)
    except Timeout:
        logger.info(f"Lock found at {lock_fp}. Returning.")
        return False

    try:
        st_time = datetime.now(tz=UTC)
        logger.info(f"Reading input dataframe from {in_fp}")
        df = read_fn(in_fp)
        logger.info("Read dataset")
        df = compute_fn(df)
        logger.info(f"Writing final output to {out_fp}")
        write_fn(df, out_fp)
        logger.info(f"Succeeded in {datetime.now(tz=UTC) - st_time}")
        return True
    finally:
        lock.release()
        try:
            lock_fp.unlink()
        except FileNotFoundError:  # pragma: no cover
            logger.warning(f"Lock file {lock_fp} was not found, though was successfully released. Returning")



"""Convert numeric values into categorical code modifiers through pre-computed or specified bins."""

import logging
import re
from collections.abc import Callable
from pathlib import Path

import polars as pl
from meds import CodeMetadataSchema, DataSchema
from omegaconf import DictConfig, OmegaConf

from ...utils import PKG_PFX, resolve_pkg_path
from .. import Stage

logger = logging.getLogger(__name__)


def _get_and_strip_format_fields(s: str) -> tuple[str, list[str]]:
    """Extracts the fields from a format string, in order.

    Args:
        s: The format string to extract fields from.

    Returns:
        A list of field names extracted from the format string.

    Examples:
        >>> _get_and_strip_format_fields("{code}//value_[{left},{right})")
        ('{}//value_[{},{})', ['code', 'left', 'right'])
        >>> _get_and_strip_format_fields("{code}")
        ('{}', ['code'])
        >>> _get_and_strip_format_fields("{code}{{{bin}}}")
        ('{}{{{}}}', ['code', 'bin'])
        >>> _get_and_strip_format_fields("code/bin")
        ('code/bin', [])
        >>> _get_and_strip_format_fields("")
        ('', [])
    """

    pattern = r"{([^{}:]*)(?::[^{}]*)?}"
    return re.sub(pattern, "{}", s), re.findall(pattern, s)


def _check_and_get_bin_endpoints(schema: pl.Schema, col: str) -> pl.Expr:
    """Checks and ensures the bin column is valid given the schema and returns the bin endpoints if so.

    Args:
        schema: The schema of the dataframe.
        col: The name of the bin column.

    Returns:
        The bin endpoints as a polars list expression.

    Raises:
        ValueError: If the bin column is not found in the schema or if it is not a struct type or if any field
            within the struct does not match the numeric value type.

    Examples:
        >>> schema = {
        ...     "numeric_value": pl.Float32,
        ...     "A": pl.Struct([pl.Field("bin1", pl.Float32), pl.Field("bin2", pl.Float32)]),
        ...     "B": pl.String,
        ...     "C": pl.Struct([pl.Field("bin1", pl.Float32), pl.Field("bin2", pl.Int64)]),
        ... }
        >>> print(_check_and_get_bin_endpoints(schema, "A"))
        .when(col("A").is_not_null()).then(col("A").multiple_fields().list.concat()).otherwise(null)
        >>> _check_and_get_bin_endpoints(schema, "B")
        Traceback (most recent call last):
            ...
        ValueError: bin_with_columns entry 'B' is not a struct type; got String.
        >>> _check_and_get_bin_endpoints(schema, "C")
        Traceback (most recent call last):
            ...
        ValueError: bin_with_columns entry 'C' has field 'bin2' with dtype Int64, which does not match the
            numeric_value dtype Float32.
        >>> _check_and_get_bin_endpoints(schema, "non_existent_col")
        Traceback (most recent call last):
            ...
        ValueError: bin_with_columns entry 'non_existent_col' not found in the dataframe schema.
    """

    if col not in schema:
        raise ValueError(f"bin_with_columns entry '{col}' not found in the dataframe schema.")

    col_dtype = schema[col]

    if not isinstance(col_dtype, pl.Struct):
        raise ValueError(f"bin_with_columns entry '{col}' is not a struct type; got {col_dtype}.")

    numeric_dtype = schema[DataSchema.numeric_value_name]

    for field in col_dtype.fields:
        if field.dtype != numeric_dtype:
            raise ValueError(
                f"bin_with_columns entry '{col}' has field '{field.name}' with dtype {field.dtype}, "
                f"which does not match the numeric_value dtype {numeric_dtype}."
            )

    return pl.when(pl.col(col).is_not_null()).then(pl.concat_list(pl.col(col).struct.unnest()))


def _get_val_bin_idx(bin_endpoints: pl.Expr) -> pl.Expr:
    """Get the bin index for the numeric value, if both it and the bin endpoints are not null.

    This is mostly separated out to enable more direct testing.

    > [!WARNING]
    > This will only work on a dataframe with a row index column named `__idx` and a numeric values column
    > named `numeric_value`.

    Args:
        bin_endpoints: The bin endpoints as a polars list expression.

    Returns:
        A polars expression that returns the bin index for the numeric value, or null if either the numeric
        value or the bin endpoints are null.

    Examples:
        >>> df = pl.DataFrame({
        ...     "numeric_value": [
        ...         -1.0, # Left of all endpoints
        ...         3.0, # Right of all endpoints
        ...         0.5, # In the middle of the first bin
        ...         1.0, # A bin endpoint directly.
        ...         None, # Null value
        ...         0.25, # Different scale, different number of bins.
        ...         1.0, # Null bin endpoints
        ...         -1.0, # Before singleton endpoint
        ...         1.0, # After singleton endpoint
        ...         0.0, # Equals singleton endpoint
        ...     ],
        ...     "bin_endpoints": [
        ...         [0.0, 1.0, 2.0],
        ...         [0.0, 1.0, 2.0],
        ...         [0.0, 1.0, 2.0],
        ...         [0.0, 1.0, 2.0],
        ...         [0.0, 1.0, 2.0],
        ...         [0.0, 0.1, 0.2, 0.3],
        ...         None,
        ...         [0.0],
        ...         [0.0],
        ...         [0.0],
        ...     ],
        ... }).with_row_index("__idx")
        >>> df
        shape: (10, 3)
        ┌───────┬───────────────┬───────────────────┐
        │ __idx ┆ numeric_value ┆ bin_endpoints     │
        │ ---   ┆ ---           ┆ ---               │
        │ u32   ┆ f64           ┆ list[f64]         │
        ╞═══════╪═══════════════╪═══════════════════╡
        │ 0     ┆ -1.0          ┆ [0.0, 1.0, 2.0]   │
        │ 1     ┆ 3.0           ┆ [0.0, 1.0, 2.0]   │
        │ 2     ┆ 0.5           ┆ [0.0, 1.0, 2.0]   │
        │ 3     ┆ 1.0           ┆ [0.0, 1.0, 2.0]   │
        │ 4     ┆ null          ┆ [0.0, 1.0, 2.0]   │
        │ 5     ┆ 0.25          ┆ [0.0, 0.1, … 0.3] │
        │ 6     ┆ 1.0           ┆ null              │
        │ 7     ┆ -1.0          ┆ [0.0]             │
        │ 8     ┆ 1.0           ┆ [0.0]             │
        │ 9     ┆ 0.0           ┆ [0.0]             │
        └───────┴───────────────┴───────────────────┘
        >>> df.with_columns(_get_val_bin_idx(pl.col("bin_endpoints")).alias("idx"))
        shape: (10, 4)
        ┌───────┬───────────────┬───────────────────┬──────┐
        │ __idx ┆ numeric_value ┆ bin_endpoints     ┆ idx  │
        │ ---   ┆ ---           ┆ ---               ┆ ---  │
        │ u32   ┆ f64           ┆ list[f64]         ┆ u32  │
        ╞═══════╪═══════════════╪═══════════════════╪══════╡
        │ 0     ┆ -1.0          ┆ [0.0, 1.0, 2.0]   ┆ 0    │
        │ 1     ┆ 3.0           ┆ [0.0, 1.0, 2.0]   ┆ 3    │
        │ 2     ┆ 0.5           ┆ [0.0, 1.0, 2.0]   ┆ 1    │
        │ 3     ┆ 1.0           ┆ [0.0, 1.0, 2.0]   ┆ 2    │
        │ 4     ┆ null          ┆ [0.0, 1.0, 2.0]   ┆ null │
        │ 5     ┆ 0.25          ┆ [0.0, 0.1, … 0.3] ┆ 3    │
        │ 6     ┆ 1.0           ┆ null              ┆ null │
        │ 7     ┆ -1.0          ┆ [0.0]             ┆ 0    │
        │ 8     ┆ 1.0           ┆ [0.0]             ┆ 1    │
        │ 9     ┆ 0.0           ┆ [0.0]             ┆ 1    │
        └───────┴───────────────┴───────────────────┴──────┘
    """

    val_col = pl.col(DataSchema.numeric_value_name)
    do_bin = bin_endpoints.is_not_null() & val_col.is_not_null()
    idx_expr = bin_endpoints.list.explode().search_sorted(val_col, side="right").over("__idx")
    return pl.when(do_bin).then(idx_expr)


def get_code(bin_endpoints: pl.Expr, val_bin_idx: pl.Expr) -> pl.Expr:
    return pl.col(DataSchema.code_name)


def get_bin(bin_endpoints: pl.Expr, val_bin_idx: pl.Expr) -> pl.Expr:
    return val_bin_idx.cast(pl.String)


def get_left(bin_endpoints: pl.Expr, val_bin_idx: pl.Expr) -> pl.Expr:
    return bin_endpoints.list.get(val_bin_idx - 1, null_on_oob=True).cast(pl.String).fill_null("-inf")


def get_right(bin_endpoints: pl.Expr, val_bin_idx: pl.Expr) -> pl.Expr:
    return bin_endpoints.list.get(val_bin_idx, null_on_oob=True).cast(pl.String).fill_null("inf")


BIN_NAME_FMT_EXPRS: dict[str, Callable[[pl.Expr, pl.Expr], pl.Expr]] = {
    "code": get_code,
    "bin": get_bin,
    "left": get_left,
    "right": get_right,
}


def add_bin_to_code(
    df: pl.LazyFrame | pl.DataFrame,
    bin_with_columns: list[str],
    code_with_bin_name: str,
    do_drop_numeric_value: bool = False,
) -> pl.LazyFrame | pl.DataFrame:
    """Converts numeric values into categorical code modifiers through a joined column of bin endpoints.

    Args:
        df: The dataframe to process. It should contain the following columns:
            - `code`: The code to be modified.
            - `numeric_value`: The numeric value to be binned.
            - All columns specified in `bin_with_columns`, which contain the linked bin endpoints to use for
              that particular measurement. Each column in `bin_with_columns` should contain a struct whose
              values are all of the same dtype and contain the increasing list of bin endpoints.
        code_with_bin_name: A string template for the code name with bin information. The template can
            leverage the keys `{code}`, `{left}`, `{right}`, and `{bin}` to capture the original code name,
            the left endpoint of the bin, the right endpoint of the bin, and the bin name.
        do_drop_numeric_value: A boolean flag indicating whether to drop the numeric_value column after
            binning. Default is `False`.

    Returns:
        A new dataframe with the codes modified to include bin information, according to the format specified.
        Codes without numeric values or with no bin endpoints included in either `code_metadata` or
        `custom_bins` are left unchanged. If `do_drop_numeric_value` is set to `True`, the returned DataFrame
        will have all cases where numeric values are converted to bins dropped. The returned dataframe will
        not have the columns contained in `bin_with_columns` in it.

    Raises:
        ValueError: If the dataframe does not contain the required columns or if the bin columns are not
            present in the schema.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 2, 3],
        ...     "code": ["lab//A", "lab//B", "lab//C", "lab//A", "lab//C", "dx//1", "lab//D"],
        ...     "numeric_value": [-1.0, 2.0, None, 1.0, 1.0, None, 1.2],
        ...     "values/quantiles": [
        ...         {"values/quantile/0.25": 0., "values/quantile/0.5": 1., "values/quantile/0.75": 2.},
        ...         {"values/quantile/0.25": -2., "values/quantile/0.5": 3., "values/quantile/0.75": 100.},
        ...         {"values/quantile/0.25": 0.01, "values/quantile/0.5": 0.4, "values/quantile/0.75": 0.6},
        ...         {"values/quantile/0.25": 0., "values/quantile/0.5": 1., "values/quantile/0.75": 2.},
        ...         {"values/quantile/0.25": 0.01, "values/quantile/0.5": 0.4, "values/quantile/0.75": 0.6},
        ...         None,
        ...         None,
        ...     ],
        ... })
        >>> df
        shape: (7, 4)
        ┌────────────┬────────┬───────────────┬──────────────────┐
        │ subject_id ┆ code   ┆ numeric_value ┆ values/quantiles │
        │ ---        ┆ ---    ┆ ---           ┆ ---              │
        │ i64        ┆ str    ┆ f64           ┆ struct[3]        │
        ╞════════════╪════════╪═══════════════╪══════════════════╡
        │ 1          ┆ lab//A ┆ -1.0          ┆ {0.0,1.0,2.0}    │
        │ 1          ┆ lab//B ┆ 2.0           ┆ {-2.0,3.0,100.0} │
        │ 1          ┆ lab//C ┆ null          ┆ {0.01,0.4,0.6}   │
        │ 2          ┆ lab//A ┆ 1.0           ┆ {0.0,1.0,2.0}    │
        │ 2          ┆ lab//C ┆ 1.0           ┆ {0.01,0.4,0.6}   │
        │ 2          ┆ dx//1  ┆ null          ┆ null             │
        │ 3          ┆ lab//D ┆ 1.2           ┆ null             │
        └────────────┴────────┴───────────────┴──────────────────┘
        >>> add_bin_to_code(df, ["values/quantiles"], "{code}//value_[{left},{right})")
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D                   ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    You can change the format to use the bin index ('{bin}') as well:

        >>> add_bin_to_code(df, ["values/quantiles"], "{code}//{bin}", do_drop_numeric_value=True)
        shape: (7, 3)
        ┌────────────┬───────────┬───────────────┐
        │ subject_id ┆ code      ┆ numeric_value │
        │ ---        ┆ ---       ┆ ---           │
        │ i64        ┆ str       ┆ f64           │
        ╞════════════╪═══════════╪═══════════════╡
        │ 1          ┆ lab//A//0 ┆ null          │
        │ 1          ┆ lab//B//1 ┆ null          │
        │ 1          ┆ lab//C    ┆ null          │
        │ 2          ┆ lab//A//2 ┆ null          │
        │ 2          ┆ lab//C//3 ┆ null          │
        │ 2          ┆ dx//1     ┆ null          │
        │ 3          ┆ lab//D    ┆ 1.2           │
        └────────────┴───────────┴───────────────┘

    You can also have multiple bin endpoint columns; in this case, the first non-null column is used. Note
    that the names of the struct keys are ignored and the structs need not match in anything but value type.

        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 1],
        ...     "code": ["A", "B", "C", "D"],
        ...     "numeric_value": [1.5, 2.5, 0.5, 0.6],
        ...     "bins_1": [
        ...         {"1": 0., "2": 1.}, None, None, {"1": 0., "2": 1.},
        ...     ],
        ...     "bins_2": [
        ...         None, None, {"a": 0.01, "b": 0.4, "c": 0.6}, {"a": -0.5, "b": 0.5, "c": 1.5},
        ...     ],
        ... })
        >>> df
        shape: (4, 5)
        ┌────────────┬──────┬───────────────┬───────────┬────────────────┐
        │ subject_id ┆ code ┆ numeric_value ┆ bins_1    ┆ bins_2         │
        │ ---        ┆ ---  ┆ ---           ┆ ---       ┆ ---            │
        │ i64        ┆ str  ┆ f64           ┆ struct[2] ┆ struct[3]      │
        ╞════════════╪══════╪═══════════════╪═══════════╪════════════════╡
        │ 1          ┆ A    ┆ 1.5           ┆ {0.0,1.0} ┆ null           │
        │ 1          ┆ B    ┆ 2.5           ┆ null      ┆ null           │
        │ 1          ┆ C    ┆ 0.5           ┆ null      ┆ {0.01,0.4,0.6} │
        │ 1          ┆ D    ┆ 0.6           ┆ {0.0,1.0} ┆ {-0.5,0.5,1.5} │
        └────────────┴──────┴───────────────┴───────────┴────────────────┘
        >>> add_bin_to_code(df, ["bins_1", "bins_2"], "[{left},{right})")
        shape: (4, 3)
        ┌────────────┬───────────┬───────────────┐
        │ subject_id ┆ code      ┆ numeric_value │
        │ ---        ┆ ---       ┆ ---           │
        │ i64        ┆ str       ┆ f64           │
        ╞════════════╪═══════════╪═══════════════╡
        │ 1          ┆ [1.0,inf) ┆ 1.5           │
        │ 1          ┆ B         ┆ 2.5           │
        │ 1          ┆ [0.4,0.6) ┆ 0.5           │
        │ 1          ┆ [0.0,1.0) ┆ 0.6           │
        └────────────┴───────────┴───────────────┘

    Errors will be raised if the dataframe is missing the numeric value field or if an invalid code name
    format string is provided (e.g., if it contains a field that is not in the list of valid fields).

        >>> add_bin_to_code(df.drop("numeric_value"), ["bins_1", "bins_2"], "{code}//{invalid_field}")
        Traceback (most recent call last):
            ...
        ValueError: Dataframe does not contain the required column 'numeric_value'.
        >>> add_bin_to_code(df, ["bins_1", "bins_2"], "{code}//{invalid_field}")
        Traceback (most recent call last):
            ...
        ValueError: Invalid bin name format field 'invalid_field' in '{code}//{invalid_field}'.
    """

    schema = df.collect_schema()

    df = df.with_row_index("__idx")

    if DataSchema.numeric_value_name not in schema:
        raise ValueError(f"Dataframe does not contain the required column '{DataSchema.numeric_value_name}'.")

    bin_endpoints = pl.coalesce([_check_and_get_bin_endpoints(schema, col) for col in bin_with_columns])
    val_bin_idx = _get_val_bin_idx(bin_endpoints)

    stripped_code_name, bin_name_fmt_cols = _get_and_strip_format_fields(code_with_bin_name)
    bin_name_fmt_exprs = []
    for n in bin_name_fmt_cols:
        if n not in BIN_NAME_FMT_EXPRS:
            raise ValueError(f"Invalid bin name format field '{n}' in '{code_with_bin_name}'.")
        bin_name_fmt_exprs.append(BIN_NAME_FMT_EXPRS[n](bin_endpoints, val_bin_idx))

    code_with_bin = pl.format(stripped_code_name, *bin_name_fmt_exprs)

    do_bin = bin_endpoints.is_not_null() & pl.col(DataSchema.numeric_value_name).is_not_null()
    df = df.with_columns(
        pl.when(do_bin).then(code_with_bin).otherwise(DataSchema.code_name).alias(DataSchema.code_name)
    )

    if do_drop_numeric_value:
        df = df.with_columns(
            pl.when(~do_bin).then(DataSchema.numeric_value_name).alias(DataSchema.numeric_value_name)
        )

    return df.drop(*bin_with_columns, "__idx")


@Stage.register
def bin_numeric_values_fntr(
    stage_cfg: DictConfig,
    code_metadata: pl.DataFrame,
    code_modifiers: list[str] | None = None,
) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Uses pre-computed value bin endpoints to add value range modifiers to code names.

    Args:
        stage_cfg: Configuration for the stage, including custom_bins and drop_numeric_value. The following
            keys are used:
                - `bin_with_columns`: A list of column names in `metadata/codes.parquet` to use for binning.
                  These columns should contain a struct whose key names correspond to bin names and whose
                  values capture the bin endpoints. The bin endpoints should correspond to the "right"
                  endpoint (exclusive) of the bin. In this way, a struct with three key-value pairs implicitly
                  creates four bins. E.g., the struct
                  `{values/quantile/0.25: 99.9, values/quantile/0.5: 105.1, values/quantile/0.75: 113.4}`
                  captures the bins `[-inf, 99.9)`, `[99.9, 105.1)`, `[105.1, 113.4)`, and `[113.4, inf)`. The
                  struct should be in sorted order of bin endpoints from least to greatest. Default is a list
                  with a single entry of `"values/quantiles"`, which captures the quantiles computed via the
                  `aggregate_code_metadata` stage. If multiple columns are specified in this list, the first
                  column that is non-null is used for each code.
                - code_with_bin_name: A string template for the code name with bin information. The template
                  can leverage the keys `{code}`, `{left}`, `{right}`, and `{bin}` to capture the original
                  code name, the left endpoint of the bin, the right endpoint of the bin, and the bin name.
                  Default is `{code}//value_[{left},{right})`.
                - do_drop_numeric_value: A boolean flag indicating whether to drop the numeric_value column
                  after binning. Default is `False`.
                - custom_bins: A dictionary mapping code names to custom bin endpoints. The keys should be the
                  code names and the values should be dictionaries with the same structure as the
                  bin_with_columns. This allows for custom binning for specific codes. Default is an empty
                  dictionary. Custom bins are used preferentially over entries in `bin_with_columns`.
                - custom_bins_filepath: Optional path to a YAML file containing custom bin endpoints. If
                  provided, the file should define the same dictionary structure as ``custom_bins``.
                  Entries loaded from this file are merged with any directly specified in ``custom_bins``.
        code_metadata: A DataFrame containing the metadata for the codes, including the bin endpoints and
            custom bins.
        code_modifiers: A list of additional columns to use for joining against codes. These columns should be
            present in both the raw data and the code metadata.

    > [!WARNING]
    > If `code_modifiers` are used, note that they do not change the structure of `custom_bins` -- that
    > strategy still only joins by `code`. File a GitHub Issue if this does not work for you.

    Returns:
        A function that takes a dataframe and returns a new dataFrame with the codes modified to include bin
        information, according to the format specified. Codes without numeric values or with no bin endpoints
        included in either `code_metadata` or `custom_bins` are left unchanged. If `do_drop_numeric_value` is
        set to `True`, the returned DataFrame will have all cases where numeric values are converted to bins
        dropped.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 2, 3],
        ...     "code": ["lab//A", "lab//B", "lab//C", "lab//A", "lab//C", "dx//1", "lab//D"],
        ...     "numeric_value": [-1.0, 2.0, None, 1.0, 1.0, None, 1.2],
        ... })
        >>> df
        shape: (7, 3)
        ┌────────────┬────────┬───────────────┐
        │ subject_id ┆ code   ┆ numeric_value │
        │ ---        ┆ ---    ┆ ---           │
        │ i64        ┆ str    ┆ f64           │
        ╞════════════╪════════╪═══════════════╡
        │ 1          ┆ lab//A ┆ -1.0          │
        │ 1          ┆ lab//B ┆ 2.0           │
        │ 1          ┆ lab//C ┆ null          │
        │ 2          ┆ lab//A ┆ 1.0           │
        │ 2          ┆ lab//C ┆ 1.0           │
        │ 2          ┆ dx//1  ┆ null          │
        │ 3          ┆ lab//D ┆ 1.2           │
        └────────────┴────────┴───────────────┘
        >>> code_metadata = pl.DataFrame({
        ...     "code": ["lab//A", "lab//B", "lab//C", "dx//1", "lab//D"],
        ...     "values/quantiles": [
        ...         {"values/quantile/0.25": 0., "values/quantile/0.5": 1., "values/quantile/0.75": 2.},
        ...         {"values/quantile/0.25": -2., "values/quantile/0.5": 3., "values/quantile/0.75": 100.},
        ...         {"values/quantile/0.25": 0.01, "values/quantile/0.5": 0.4, "values/quantile/0.75": 0.6},
        ...         None,
        ...         None,
        ...     ],
        ... })

    If we run the functor from a default (empty) stage configuration, we get the following:

        >>> fn = bin_numeric_values_fntr(DictConfig({}), code_metadata)
        >>> fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D                   ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    We can use the stage configuration to change the format of the bin name...

        >>> fn = bin_numeric_values_fntr(DictConfig({"code_with_bin_name": "{code}//{bin}"}), code_metadata)
        >>> fn(df)
        shape: (7, 3)
        ┌────────────┬───────────┬───────────────┐
        │ subject_id ┆ code      ┆ numeric_value │
        │ ---        ┆ ---       ┆ ---           │
        │ i64        ┆ str       ┆ f64           │
        ╞════════════╪═══════════╪═══════════════╡
        │ 1          ┆ lab//A//0 ┆ -1.0          │
        │ 1          ┆ lab//B//1 ┆ 2.0           │
        │ 1          ┆ lab//C    ┆ null          │
        │ 2          ┆ lab//A//2 ┆ 1.0           │
        │ 2          ┆ lab//C//3 ┆ 1.0           │
        │ 2          ┆ dx//1     ┆ null          │
        │ 3          ┆ lab//D    ┆ 1.2           │
        └────────────┴───────────┴───────────────┘

    Drop the numeric values we bin:

        >>> fn = bin_numeric_values_fntr(DictConfig({"drop_numeric_value": True}), code_metadata)
        >>> fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ null          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ null          │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ null          │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ null          │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D                   ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    Add custom bin endpoints:

        >>> fn = bin_numeric_values_fntr(DictConfig({"custom_bins": {"lab//D": {"foo": 1.0}}}), code_metadata)
        >>> fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D//value_[1.0,inf)  ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    Load custom bins from a YAML file on disk:

        >>> import tempfile, yaml
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     bins_fp = Path(tmpdir) / "bins.yaml"
        ...     yaml.safe_dump({"lab//D": {"foo": 1.0}}, bins_fp.open("w"))
        ...     fn = bin_numeric_values_fntr(
        ...         DictConfig({"custom_bins_filepath": str(bins_fp)}),
        ...         code_metadata,
        ...     )
        ...     fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D//value_[1.0,inf)  ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    If the filepath does not exist or the YAML does not contain a dictionary,
    errors will be raised:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     fn = bin_numeric_values_fntr(
        ...         DictConfig({"custom_bins_filepath": str(Path(tmpdir) / "missing.yaml")}),
        ...         code_metadata,
        ...     )
        Traceback (most recent call last):
            ...
        FileNotFoundError: custom_bins_filepath '/tmp/tmp.../missing.yaml' does not exist.
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     bins_fp = Path(tmpdir) / "bins.yaml"
        ...     yaml.safe_dump([1, 2], bins_fp.open("w"))
        ...     bin_numeric_values_fntr(
        ...         DictConfig({"custom_bins_filepath": str(bins_fp)}),
        ...         code_metadata,
        ...     )
        Traceback (most recent call last):
            ...
        TypeError: custom_bins_filepath must point to a YAML file with a dictionary

    Package paths can also be used when patched with ``resolve_pkg_path``:

        >>> from unittest.mock import patch
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as bins_yaml:
        ...     bins_fp = Path(bins_yaml.name)
        ...     yaml.safe_dump({"lab//D": {"foo": 1.0}}, bins_fp.open("w"))
        ...     with patch(
        ...         "MEDS_transforms.stages.bin_numeric_values.bin_numeric_values.resolve_pkg_path",
        ...         return_value=bins_fp,
        ...     ):
        ...         fn = bin_numeric_values_fntr(
        ...             DictConfig({"custom_bins_filepath": "pkg://fake_pkg.bins.yaml"}),
        ...             code_metadata,
        ...         )
        ...         fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D//value_[1.0,inf)  ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    Use different bin columns (sourced from the code metadata)

        >>> code_metadata = pl.DataFrame({
        ...     "code": ["lab//A", "lab//B", "lab//C", "dx//1", "lab//D"],
        ...     "bins_1": [
        ...         {"values/quantile/0.25": 0., "values/quantile/0.5": 1., "values/quantile/0.75": 2.},
        ...         None,
        ...         {"values/quantile/0.25": 0.01, "values/quantile/0.5": 0.4, "values/quantile/0.75": 0.6},
        ...         None,
        ...         None,
        ...     ],
        ...     "bins_2": [
        ...         {"values/quantile/0.25": -2., "values/quantile/0.5": 3., "values/quantile/0.75": 100.},
        ...         {"values/quantile/0.25": -2., "values/quantile/0.5": 3., "values/quantile/0.75": 100.},
        ...         None,
        ...         None,
        ...         None,
        ...     ],
        ... })
        >>> fn = bin_numeric_values_fntr(
        ...     DictConfig({"bin_with_columns": ["bins_1", "bins_2"]}), code_metadata
        ... )
        >>> fn(df)
        shape: (7, 3)
        ┌────────────┬──────────────────────────┬───────────────┐
        │ subject_id ┆ code                     ┆ numeric_value │
        │ ---        ┆ ---                      ┆ ---           │
        │ i64        ┆ str                      ┆ f64           │
        ╞════════════╪══════════════════════════╪═══════════════╡
        │ 1          ┆ lab//A//value_[-inf,0.0) ┆ -1.0          │
        │ 1          ┆ lab//B//value_[-2.0,3.0) ┆ 2.0           │
        │ 1          ┆ lab//C                   ┆ null          │
        │ 2          ┆ lab//A//value_[1.0,2.0)  ┆ 1.0           │
        │ 2          ┆ lab//C//value_[0.6,inf)  ┆ 1.0           │
        │ 2          ┆ dx//1                    ┆ null          │
        │ 3          ┆ lab//D                   ┆ 1.2           │
        └────────────┴──────────────────────────┴───────────────┘

    We can also use code modifiers to join in a richer manner against the code metadata:

        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 2, 3],
        ...     "code": ["lab", "lab", "lab", "lab", "lab", "dx//1", "lab"],
        ...     "unit": ["A", "B", "C", "A", "C", None, "D"],
        ...     "numeric_value": [-1.0, 2.0, None, 1.0, 1.0, None, 1.2],
        ... })
        >>> df
        shape: (7, 4)
        ┌────────────┬───────┬──────┬───────────────┐
        │ subject_id ┆ code  ┆ unit ┆ numeric_value │
        │ ---        ┆ ---   ┆ ---  ┆ ---           │
        │ i64        ┆ str   ┆ str  ┆ f64           │
        ╞════════════╪═══════╪══════╪═══════════════╡
        │ 1          ┆ lab   ┆ A    ┆ -1.0          │
        │ 1          ┆ lab   ┆ B    ┆ 2.0           │
        │ 1          ┆ lab   ┆ C    ┆ null          │
        │ 2          ┆ lab   ┆ A    ┆ 1.0           │
        │ 2          ┆ lab   ┆ C    ┆ 1.0           │
        │ 2          ┆ dx//1 ┆ null ┆ null          │
        │ 3          ┆ lab   ┆ D    ┆ 1.2           │
        └────────────┴───────┴──────┴───────────────┘
        >>> code_metadata = pl.DataFrame({
        ...     "code": ["lab", "lab", "lab", "dx//1", "lab"],
        ...     "unit": ["A", "B", "C", None, "D"],
        ...     "values/quantiles": [
        ...         {"values/quantile/0.25": 0., "values/quantile/0.5": 1., "values/quantile/0.75": 2.},
        ...         {"values/quantile/0.25": -2., "values/quantile/0.5": 3., "values/quantile/0.75": 100.},
        ...         {"values/quantile/0.25": 0.01, "values/quantile/0.5": 0.4, "values/quantile/0.75": 0.6},
        ...         None,
        ...         None,
        ...     ],
        ... })
        >>> fn = bin_numeric_values_fntr(DictConfig({}), code_metadata, ["unit"])
        >>> fn(df)
        shape: (7, 4)
        ┌────────────┬───────────────────────┬──────┬───────────────┐
        │ subject_id ┆ code                  ┆ unit ┆ numeric_value │
        │ ---        ┆ ---                   ┆ ---  ┆ ---           │
        │ i64        ┆ str                   ┆ str  ┆ f64           │
        ╞════════════╪═══════════════════════╪══════╪═══════════════╡
        │ 1          ┆ lab//value_[-inf,0.0) ┆ A    ┆ -1.0          │
        │ 1          ┆ lab//value_[-2.0,3.0) ┆ B    ┆ 2.0           │
        │ 1          ┆ lab                   ┆ C    ┆ null          │
        │ 2          ┆ lab//value_[1.0,2.0)  ┆ A    ┆ 1.0           │
        │ 2          ┆ lab//value_[0.6,inf)  ┆ C    ┆ 1.0           │
        │ 2          ┆ dx//1                 ┆ null ┆ null          │
        │ 3          ┆ lab                   ┆ D    ┆ 1.2           │
        └────────────┴───────────────────────┴──────┴───────────────┘
    """
    if code_modifiers is None:
        code_modifiers = []

    # Step 1: Add custom_bins column to code_metadata
    custom_bins = stage_cfg.get("custom_bins", {})
    custom_bins_fp = stage_cfg.get("custom_bins_filepath")

    if isinstance(custom_bins, DictConfig):
        custom_bins = OmegaConf.to_container(custom_bins)

    if custom_bins_fp:
        fp = Path(custom_bins_fp)
        if custom_bins_fp.startswith(PKG_PFX):
            fp = resolve_pkg_path(custom_bins_fp)
        if not fp.is_file():
            raise FileNotFoundError(f"custom_bins_filepath '{custom_bins_fp}' does not exist.")

        file_bins = OmegaConf.load(fp)
        if isinstance(file_bins, DictConfig):
            file_bins = OmegaConf.to_container(file_bins)
        if not isinstance(file_bins, dict):
            raise TypeError("custom_bins_filepath must point to a YAML file with a dictionary")
        custom_bins = {**file_bins, **custom_bins}

    do_use_custom_bins = bool(custom_bins)
    do_drop_numeric_value = stage_cfg.get("drop_numeric_value", False)
    bin_with_columns = stage_cfg.get("bin_with_columns", ["values/quantiles"])
    code_with_bin_name = stage_cfg.get("code_with_bin_name", "{code}//value_[{left},{right})")

    if do_use_custom_bins:
        if isinstance(custom_bins, DictConfig):
            custom_bins = OmegaConf.to_container(custom_bins)

        struct_dtype = pl.Struct(dict.fromkeys(next(iter(custom_bins.values())).keys(), pl.Float32))
        custom_bins_series = pl.Series(
            [custom_bins.get(c, None) for c in code_metadata[CodeMetadataSchema.code_name]],
            dtype=struct_dtype,
        )
        code_metadata = code_metadata.with_columns(custom_bins_series.alias("__custom_bins"))
        bin_with_columns = ["__custom_bins", *bin_with_columns]

    join_cols = [CodeMetadataSchema.code_name, *code_modifiers]

    code_metadata = code_metadata.select(*join_cols, *bin_with_columns)

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        numeric_dtype = df.collect_schema()[DataSchema.numeric_value_name]

        cast_exprs = []
        for col in bin_with_columns:
            new_struct_dtype = pl.Struct({f.name: numeric_dtype for f in code_metadata.schema[col].fields})
            cast_exprs.append(pl.col(col).cast(new_struct_dtype))

        local_metadata = code_metadata.with_columns(*cast_exprs)

        if isinstance(df, pl.LazyFrame):
            local_metadata = local_metadata.lazy()

        df = df.join(local_metadata, on=join_cols, how="left", maintain_order="left")
        return add_bin_to_code(df, bin_with_columns, code_with_bin_name, do_drop_numeric_value)

    return fn



"""Utilities to help add time-derived measurements (e.g., a subject's age) to a MEDS dataset."""

import logging

import polars as pl
from meds import DataSchema

logger = logging.getLogger(__name__)

TIME_DURATION_UNITS = {
    "seconds": (["s", "sec", "secs", "second", "seconds"], 1),
    "minutes": (["m", "min", "mins", "minute", "minutes"], 60),
    "hours": (["h", "hr", "hrs", "hour", "hours"], 60 * 60),
    "days": (["d", "day", "days"], 60 * 60 * 24),
    "weeks": (["w", "wk", "wks", "week", "weeks"], 60 * 60 * 24 * 7),
    "months": (["mo", "mos", "month", "months"], 60 * 60 * 24 * 30.436875),
    "years": (["y", "yr", "yrs", "year", "years"], 60 * 60 * 24 * 365.2422),
}


def normalize_time_unit(unit: str) -> tuple[str, float]:
    """Normalize a time unit string to a canonical form and return the number of seconds in that unit.

    Note that this function is designed for computing _approximate_ time durations over long periods, not
    canonical, local calendar time durations. E.g., a "month" is not a fixed number of seconds, but this
    function will return the average number of seconds in a month, accounting for leap years.

    TODO: consider replacing this function with the use of https://github.com/wroberts/pytimeparse

    Args:
        unit: The input unit to normalize.

    Returns:
        A tuple containing the canonical unit and the number of seconds in that unit.

    Raises:
        ValueError: If the input unit is not recognized.

    Examples:
        >>> normalize_time_unit("s")
        ('seconds', 1)
        >>> normalize_time_unit("min")
        ('minutes', 60)
        >>> normalize_time_unit("hours")
        ('hours', 3600)
        >>> normalize_time_unit("day")
        ('days', 86400)
        >>> normalize_time_unit("wks")
        ('weeks', 604800)
        >>> normalize_time_unit("month")
        ('months', 2629746.0)
        >>> normalize_time_unit("years")
        ('years', 31556926.080000002)
        >>> normalize_time_unit("fortnight")
        Traceback (most recent call last):
            ...
        ValueError: Unknown time unit 'fortnight'. Valid units include:
          * seconds: s, sec, secs, second, seconds
          * minutes: m, min, mins, minute, minutes
          * hours: h, hr, hrs, hour, hours
          * days: d, day, days
          * weeks: w, wk, wks, week, weeks
          * months: mo, mos, month, months
          * years: y, yr, yrs, year, years
    """
    for canonical_unit, (aliases, seconds) in TIME_DURATION_UNITS.items():
        if unit in aliases:
            return canonical_unit, seconds

    valid_unit_lines = []
    for canonical, (aliases, _) in TIME_DURATION_UNITS.items():
        valid_unit_lines.append(f"  * {canonical}: {', '.join(aliases)}")
    valid_units_str = "\n".join(valid_unit_lines)
    raise ValueError(f"Unknown time unit '{unit}'. Valid units include:\n{valid_units_str}")


def unique_events(df: pl.DataFrame | pl.LazyFrame) -> pl.DataFrame | pl.LazyFrame:
    """Returns rows corresponding to unique events (time-points) in a MEDS dataframe.

    Args:
        df: The input DataFrame.

    Returns:
        A DataFrame containing rows with unique time-points for each subject, with only subject ID and time
        columns included.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 2, 2],
        ...         "time": [
        ...             None,
        ...             datetime(2021, 1, 1, 12, 0),
        ...             datetime(2021, 1, 1, 12, 0),
        ...             datetime(2022, 1, 1, 18, 0),
        ...             datetime(2022, 1, 1, 18, 0),
        ...             datetime(2022, 1, 1, 18, 0),
        ...         ],
        ...         "code": ["static", "lab//A", "lab//A", "lab//B", "lab//B", "dx//1"],
        ...     },
        ... )
        >>> df
        shape: (6, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ i64        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 2021-01-01 12:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-01 12:00:00 ┆ lab//A │
        │ 1          ┆ 2022-01-01 18:00:00 ┆ lab//B │
        │ 2          ┆ 2022-01-01 18:00:00 ┆ lab//B │
        │ 2          ┆ 2022-01-01 18:00:00 ┆ dx//1  │
        └────────────┴─────────────────────┴────────┘
        >>> unique_events(df)
        shape: (3, 2)
        ┌────────────┬─────────────────────┐
        │ subject_id ┆ time                │
        │ ---        ┆ ---                 │
        │ i64        ┆ datetime[μs]        │
        ╞════════════╪═════════════════════╡
        │ 1          ┆ 2021-01-01 12:00:00 │
        │ 1          ┆ 2022-01-01 18:00:00 │
        │ 2          ┆ 2022-01-01 18:00:00 │
        └────────────┴─────────────────────┘
    """

    return (
        df.drop_nulls(subset=[DataSchema.time_name])
        .select(DataSchema.subject_id_name, DataSchema.time_name)
        .unique(maintain_order=True)
    )



"""Generate documentation pages for built-in stages using mkdocs_gen_files."""

from pathlib import Path

import mkdocs_gen_files

from MEDS_transforms.stages.docgen import generate_stage_docs

nav = mkdocs_gen_files.Nav()

for doc in generate_stage_docs("MEDS_transforms"):
    nav_key = Path(doc.path)
    doc_path = nav_key / "index.md"
    full_doc_path = Path("stages") / doc_path

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        fd.write(doc.content)

    nav[nav_key.parts] = doc_path.as_posix()

    if doc.edit_path:
        mkdocs_gen_files.set_edit_path(doc.path, doc.edit_path)

with mkdocs_gen_files.open("stages/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



import logging
import sys
from importlib.resources import files

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf

from . import __package_name__, __version__
from .configs import PipelineConfig
from .stages.discovery import get_all_registered_stages

logger = logging.getLogger(__name__)

HELP_STRS = {"--help", "-h", "help", "h"}
MAIN_CFG_PATH = files(__package_name__) / "configs" / "_main.yaml"


def print_help_stage():
    """Print help for all stages."""

    all_stage_names = list(get_all_registered_stages().keys())

    print(f"Usage: {sys.argv[0]} <pipeline_yaml> <stage_name> [args]")
    print(
        "  * pipeline_yaml: Path to the pipeline YAML file on disk or in the "
        "'pkg://<pkg_name>.<relative_path>' format."
    )
    print("  * stage_name: Name of the stage to run.")
    print()
    print("Available stages:")
    for name in sorted(all_stage_names):
        print(f"  - {name}")


def run_stage():
    """Run a stage based on command line arguments."""

    if len(sys.argv) < 2:
        print_help_stage()
        sys.exit(1)
    elif sys.argv[1] in HELP_STRS:
        print_help_stage()
        sys.exit(0)
    elif len(sys.argv) < 3:
        print_help_stage()
        sys.exit(1)

    pipeline_cfg = PipelineConfig.from_arg(sys.argv[1])
    stage_name = sys.argv[2]

    sys.argv = sys.argv[2:]  # remove dispatcher arguments

    # Register the stage structured config and pipeline configuration
    stage = pipeline_cfg.register_for(stage_name)

    cs = ConfigStore.instance()
    cs.store(name="_main", node=OmegaConf.load(MAIN_CFG_PATH))

    hydra_wrapper = hydra.main(version_base=None, config_name="_main")

    OmegaConf.register_new_resolver("get_package_version", lambda: __version__, replace=False)
    OmegaConf.register_new_resolver("get_package_name", lambda: __package_name__, replace=False)
    OmegaConf.register_new_resolver("stage_name", lambda: stage_name)
    OmegaConf.register_new_resolver("stage_docstring", lambda: stage.stage_docstring.replace("$", "$$"))

    hydra_wrapper(stage.main)()



"""Basic code for a mapreduce stage."""

import logging
from datetime import UTC, datetime
from functools import partial
from pathlib import Path

import polars as pl
from meds import DataSchema, SubjectSplitSchema, subject_splits_filepath
from omegaconf import DictConfig

from ..compute_modes import (
    ANY_COMPUTE_FN_T,
    COMPUTE_FN_T,
    bind_compute_fn,
    is_match_revise,
    match_revise_fntr,
)
from ..dataframe import DF_T, READ_FN_T, WRITE_FN_T, read_and_filter_fntr, read_df, write_df
from .mapper import map_over
from .reducer import REDUCE_FN_T, reduce_over
from .shard_iteration import SHARD_ITR_FNTR_T, shard_iterator

logger = logging.getLogger(__name__)


def map_stage(
    cfg: DictConfig,
    map_fn: COMPUTE_FN_T,
    read_fn: READ_FN_T | None = None,
    write_fn: WRITE_FN_T | None = None,
    shard_iterator_fntr: SHARD_ITR_FNTR_T = shard_iterator,
) -> list[Path]:
    """Performs a mapping stage operation on shards produced by the shard iterator.

    Args:
        cfg: Configuration dictionary containing stage_cfg, input_dir, and other necessary parameters.
        map_fn: Function to apply to each shard.
        read_fn: Function to read data from the input file paths. Defaults to reading parquet files with
            polars.
        write_fn: Function to write the transformed data to the output file paths. Defaults to writing
            parquet files with polars.
        shard_iterator_fntr: Function to create the shard iterator. Defaults to the default shard iterator,
            which iterates over the parquet files in the data input directory within the `cfg.stage_cfg` in a
            pseudo-random, worker-dependent manner.

    Returns:
        List of output file paths that were written.

    Raises:
        FileNotFoundError: If train_only is True but subject split file doesn't exist.
        ValueError: If includes_only_train is True but train_only is False.

    Examples:
        >>> from meds_testing_helpers.dataset import MEDSDataset
        >>> D = MEDSDataset(root_dir=simple_static_MEDS)

    We'll show an example of this function using the `simple_static_MEDS` dataset provided as a pytest
    fixture and loaded in this doctest via our `conftest.py` file in the
    [`meds_testing_helpers`](https://github.com/Medical-Event-Data-Standard/meds_testing_helpers) package.

    To see this data, let's inspect it, first on disk (which is merely the typical MEDS fashion):

        >>> print_directory(simple_static_MEDS)
        ├── data
        │   ├── held_out
        │   │   └── 0.parquet
        │   ├── train
        │   │   ├── 0.parquet
        │   │   └── 1.parquet
        │   └── tuning
        │       └── 0.parquet
        └── metadata
            ├── codes.parquet
            ├── dataset.json
            └── subject_splits.parquet

    As well as its actual contents:

        >>> for k, df in D._pl_shards.items():
        ...     print(f"{k}:")
        ...     print(df)
        held_out/0:
        shape: (11, 4)
        ┌────────────┬─────────────────────┬───────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                  ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                   ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                   ┆ f32           │
        ╞════════════╪═════════════════════╪═══════════════════════╪═══════════════╡
        │ 1500733    ┆ null                ┆ EYE_COLOR//BROWN      ┆ null          │
        │ 1500733    ┆ null                ┆ HEIGHT                ┆ 158.601318    │
        │ 1500733    ┆ 1986-07-20 00:00:00 ┆ DOB                   ┆ null          │
        │ 1500733    ┆ 2010-06-03 14:54:38 ┆ ADMISSION//ORTHOPEDIC ┆ null          │
        │ 1500733    ┆ 2010-06-03 14:54:38 ┆ HR                    ┆ 91.400002     │
        │ …          ┆ …                   ┆ …                     ┆ …             │
        │ 1500733    ┆ 2010-06-03 15:39:49 ┆ HR                    ┆ 84.400002     │
        │ 1500733    ┆ 2010-06-03 15:39:49 ┆ TEMP                  ┆ 100.300003    │
        │ 1500733    ┆ 2010-06-03 16:20:49 ┆ HR                    ┆ 90.099998     │
        │ 1500733    ┆ 2010-06-03 16:20:49 ┆ TEMP                  ┆ 100.099998    │
        │ 1500733    ┆ 2010-06-03 16:44:26 ┆ DISCHARGE             ┆ null          │
        └────────────┴─────────────────────┴───────────────────────┴───────────────┘
        train/0:
        shape: (30, 4)
        ┌────────────┬─────────────────────┬────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code               ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                ┆ f32           │
        ╞════════════╪═════════════════════╪════════════════════╪═══════════════╡
        │ 239684     ┆ null                ┆ EYE_COLOR//BROWN   ┆ null          │
        │ 239684     ┆ null                ┆ HEIGHT             ┆ 175.271118    │
        │ 239684     ┆ 1980-12-28 00:00:00 ┆ DOB                ┆ null          │
        │ 239684     ┆ 2010-05-11 17:41:51 ┆ ADMISSION//CARDIAC ┆ null          │
        │ 239684     ┆ 2010-05-11 17:41:51 ┆ HR                 ┆ 102.599998    │
        │ …          ┆ …                   ┆ …                  ┆ …             │
        │ 1195293    ┆ 2010-06-20 20:24:44 ┆ HR                 ┆ 107.699997    │
        │ 1195293    ┆ 2010-06-20 20:24:44 ┆ TEMP               ┆ 100.0         │
        │ 1195293    ┆ 2010-06-20 20:41:33 ┆ HR                 ┆ 107.5         │
        │ 1195293    ┆ 2010-06-20 20:41:33 ┆ TEMP               ┆ 100.400002    │
        │ 1195293    ┆ 2010-06-20 20:50:04 ┆ DISCHARGE          ┆ null          │
        └────────────┴─────────────────────┴────────────────────┴───────────────┘
        train/1:
        shape: (14, 4)
        ┌────────────┬─────────────────────┬───────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                  ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                   ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                   ┆ f32           │
        ╞════════════╪═════════════════════╪═══════════════════════╪═══════════════╡
        │ 68729      ┆ null                ┆ EYE_COLOR//HAZEL      ┆ null          │
        │ 68729      ┆ null                ┆ HEIGHT                ┆ 160.395309    │
        │ 68729      ┆ 1978-03-09 00:00:00 ┆ DOB                   ┆ null          │
        │ 68729      ┆ 2010-05-26 02:30:56 ┆ ADMISSION//PULMONARY  ┆ null          │
        │ 68729      ┆ 2010-05-26 02:30:56 ┆ HR                    ┆ 86.0          │
        │ …          ┆ …                   ┆ …                     ┆ …             │
        │ 814703     ┆ 1976-03-28 00:00:00 ┆ DOB                   ┆ null          │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ ADMISSION//ORTHOPEDIC ┆ null          │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ HR                    ┆ 170.199997    │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ TEMP                  ┆ 100.099998    │
        │ 814703     ┆ 2010-02-05 07:02:30 ┆ DISCHARGE             ┆ null          │
        └────────────┴─────────────────────┴───────────────────────┴───────────────┘
        tuning/0:
        shape: (7, 4)
        ┌────────────┬─────────────────────┬──────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                 ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                  ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                  ┆ f32           │
        ╞════════════╪═════════════════════╪══════════════════════╪═══════════════╡
        │ 754281     ┆ null                ┆ EYE_COLOR//BROWN     ┆ null          │
        │ 754281     ┆ null                ┆ HEIGHT               ┆ 166.22261     │
        │ 754281     ┆ 1988-12-19 00:00:00 ┆ DOB                  ┆ null          │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ ADMISSION//PULMONARY ┆ null          │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ HR                   ┆ 142.0         │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ TEMP                 ┆ 99.800003     │
        │ 754281     ┆ 2010-01-03 08:22:13 ┆ DISCHARGE            ┆ null          │
        └────────────┴─────────────────────┴──────────────────────┴───────────────┘
        >>> D._pl_subject_splits
        shape: (6, 2)
        ┌────────────┬──────────┐
        │ subject_id ┆ split    │
        │ ---        ┆ ---      │
        │ i64        ┆ str      │
        ╞════════════╪══════════╡
        │ 239684     ┆ train    │
        │ 1195293    ┆ train    │
        │ 68729      ┆ train    │
        │ 814703     ┆ train    │
        │ 754281     ┆ tuning   │
        │ 1500733    ┆ held_out │
        └────────────┴──────────┘

    We'll also make a simple helper too to print the output for us

        >>> def profile_map_stage(test_dir: str, in_MEDS_dir: Path | None = None, **kwargs):
        ...     '''Makes a test config, adds the output directory, runs the mapping stage, & shows outputs'''
        ...     test_cfg = DictConfig(cfg)
        ...     test_cfg.stage_cfg.output_dir = Path(test_dir)
        ...     if in_MEDS_dir is not None:
        ...         test_cfg.input_dir = str(in_MEDS_dir)
        ...         test_cfg.stage_cfg.data_input_dir = str(in_MEDS_dir / "data")
        ...         test_cfg.stage_cfg.metadata_input_dir = str(in_MEDS_dir / "metadata")
        ...     out_fps = map_stage(cfg=test_cfg, **kwargs)
        ...     print("Output directory:")
        ...     print_directory(test_cfg.stage_cfg.output_dir)
        ...     print("------------------")
        ...     print("Output files:")
        ...     print("------------------")
        ...     for fp in out_fps:
        ...         print(f"  - {fp.relative_to(test_cfg.stage_cfg.output_dir)}:")
        ...         print(pl.read_parquet(fp))

    To map over this data, we need a configuration file that will point our data input dir to this dataset.
    Normally, the stage configuration is handled automatically by the Stage objects and the `PipelineConfig`
    class, but we'll just fudge it here for the sake of the example. Note we haven't added an output dir yet
    -- that will be a temporary directory we'll create just before we run the stage.

        >>> cfg = DictConfig(
        ...     {
        ...         "worker": 0,
        ...         "do_overwrite": False,
        ...         "input_dir": str(simple_static_MEDS),
        ...         "stage_cfg": {
        ...             "data_input_dir": str(simple_static_MEDS / "data"),
        ...             "metadata_input_dir": str(simple_static_MEDS / "metadata"),
        ...             "output_dir": "???",
        ...         },
        ...     }
        ... )

    We'll also need a mapping function that will be applied to each shard. For this example, we'll just count
    the number of occurrences of each code:

        >>> def count_codes(df: pl.LazyFrame) -> pl.LazyFrame:
        ...     return df.group_by("code", maintain_order=True).agg(pl.len().alias("count"))

    Now we can run the mapping stage:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     profile_map_stage(tmpdir, map_fn=count_codes)
        Output directory:
        ├── held_out
        │   └── 0.parquet
        ├── train
        │   ├── 0.parquet
        │   └── 1.parquet
        └── tuning
            └── 0.parquet
        ------------------
        Output files:
        ------------------
          - held_out/0.parquet:
        shape: (7, 2)
        ┌───────────────────────┬───────┐
        │ code                  ┆ count │
        │ ---                   ┆ ---   │
        │ str                   ┆ u32   │
        ╞═══════════════════════╪═══════╡
        │ EYE_COLOR//BROWN      ┆ 1     │
        │ HEIGHT                ┆ 1     │
        │ DOB                   ┆ 1     │
        │ ADMISSION//ORTHOPEDIC ┆ 1     │
        │ HR                    ┆ 3     │
        │ TEMP                  ┆ 3     │
        │ DISCHARGE             ┆ 1     │
        └───────────────────────┴───────┘
          - tuning/0.parquet:
        shape: (7, 2)
        ┌──────────────────────┬───────┐
        │ code                 ┆ count │
        │ ---                  ┆ ---   │
        │ str                  ┆ u32   │
        ╞══════════════════════╪═══════╡
        │ EYE_COLOR//BROWN     ┆ 1     │
        │ HEIGHT               ┆ 1     │
        │ DOB                  ┆ 1     │
        │ ADMISSION//PULMONARY ┆ 1     │
        │ HR                   ┆ 1     │
        │ TEMP                 ┆ 1     │
        │ DISCHARGE            ┆ 1     │
        └──────────────────────┴───────┘
          - train/1.parquet:
        shape: (8, 2)
        ┌───────────────────────┬───────┐
        │ code                  ┆ count │
        │ ---                   ┆ ---   │
        │ str                   ┆ u32   │
        ╞═══════════════════════╪═══════╡
        │ EYE_COLOR//HAZEL      ┆ 2     │
        │ HEIGHT                ┆ 2     │
        │ DOB                   ┆ 2     │
        │ ADMISSION//PULMONARY  ┆ 1     │
        │ HR                    ┆ 2     │
        │ TEMP                  ┆ 2     │
        │ DISCHARGE             ┆ 2     │
        │ ADMISSION//ORTHOPEDIC ┆ 1     │
        └───────────────────────┴───────┘
          - train/0.parquet:
        shape: (8, 2)
        ┌────────────────────┬───────┐
        │ code               ┆ count │
        │ ---                ┆ ---   │
        │ str                ┆ u32   │
        ╞════════════════════╪═══════╡
        │ EYE_COLOR//BROWN   ┆ 1     │
        │ HEIGHT             ┆ 2     │
        │ DOB                ┆ 2     │
        │ ADMISSION//CARDIAC ┆ 2     │
        │ HR                 ┆ 10    │
        │ TEMP               ┆ 10    │
        │ DISCHARGE          ┆ 2     │
        │ EYE_COLOR//BLUE    ┆ 1     │
        └────────────────────┴───────┘

    If we set `train_only` to `True`, we can see that the held-out and tuning splits are not included in the
    output:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     cfg.stage_cfg.train_only = True
        ...     profile_map_stage(tmpdir, map_fn=count_codes)
        ...     cfg.stage_cfg.train_only = False
        Output directory:
        └── train
            ├── 0.parquet
            └── 1.parquet
        ------------------
        Output files:
        ------------------
          - train/1.parquet:
        shape: (8, 2)
        ┌───────────────────────┬───────┐
        │ code                  ┆ count │
        │ ---                   ┆ ---   │
        │ str                   ┆ u32   │
        ╞═══════════════════════╪═══════╡
        │ EYE_COLOR//HAZEL      ┆ 2     │
        │ HEIGHT                ┆ 2     │
        │ DOB                   ┆ 2     │
        │ ADMISSION//PULMONARY  ┆ 1     │
        │ HR                    ┆ 2     │
        │ TEMP                  ┆ 2     │
        │ DISCHARGE             ┆ 2     │
        │ ADMISSION//ORTHOPEDIC ┆ 1     │
        └───────────────────────┴───────┘
          - train/0.parquet:
        shape: (8, 2)
        ┌────────────────────┬───────┐
        │ code               ┆ count │
        │ ---                ┆ ---   │
        │ str                ┆ u32   │
        ╞════════════════════╪═══════╡
        │ EYE_COLOR//BROWN   ┆ 1     │
        │ HEIGHT             ┆ 2     │
        │ DOB                ┆ 2     │
        │ ADMISSION//CARDIAC ┆ 2     │
        │ HR                 ┆ 10    │
        │ TEMP               ┆ 10    │
        │ DISCHARGE          ┆ 2     │
        │ EYE_COLOR//BLUE    ┆ 1     │
        └────────────────────┴───────┘

    Note that, here, the held-out and tuning shards outright are not included in the output, as files or
    otherwise -- this is because the `shard_iterator` function can tell by the prefix that this dataset is
    sharded by split. What if that weren't the case? To show that, we need to copy our MEDS dataset to a form
    not sharded by split. We'll use a simple helper to do that.

        >>> import shutil
        >>> def copy_MEDS_without_split_sharding(output_dir: Path):
        ...     '''Copy the MEDS dataset to a new location without split sharding.'''
        ...     for fp in Path(simple_static_MEDS).rglob("*.*"):
        ...         relative_path = fp.relative_to(simple_static_MEDS)
        ...         if relative_path.parts[0] == "data":
        ...             # Remove the {split}/ prefix
        ...             relative_path = Path("data/") / "_".join(relative_path.parts[1:])
        ...         out_fp = output_dir / relative_path
        ...         out_fp.parent.mkdir(parents=True, exist_ok=True)
        ...         shutil.copy(fp, out_fp)
        >>> with tempfile.TemporaryDirectory() as MEDS_dir, tempfile.TemporaryDirectory() as out_dir:
        ...     copy_MEDS_without_split_sharding(Path(MEDS_dir))
        ...     cfg.stage_cfg.train_only = True
        ...     profile_map_stage(out_dir, in_MEDS_dir=Path(MEDS_dir), map_fn=count_codes)
        ...     cfg.stage_cfg.train_only = False
        Output directory:
        ├── held_out_0.parquet
        ├── train_0.parquet
        ├── train_1.parquet
        └── tuning_0.parquet
        ------------------
        Output files:
        ------------------
          - train_1.parquet:
        shape: (8, 2)
        ┌───────────────────────┬───────┐
        │ code                  ┆ count │
        │ ---                   ┆ ---   │
        │ str                   ┆ u32   │
        ╞═══════════════════════╪═══════╡
        │ EYE_COLOR//HAZEL      ┆ 2     │
        │ HEIGHT                ┆ 2     │
        │ DOB                   ┆ 2     │
        │ ADMISSION//PULMONARY  ┆ 1     │
        │ HR                    ┆ 2     │
        │ TEMP                  ┆ 2     │
        │ DISCHARGE             ┆ 2     │
        │ ADMISSION//ORTHOPEDIC ┆ 1     │
        └───────────────────────┴───────┘
          - tuning_0.parquet:
        shape: (0, 2)
        ┌──────┬───────┐
        │ code ┆ count │
        │ ---  ┆ ---   │
        │ str  ┆ u32   │
        ╞══════╪═══════╡
        └──────┴───────┘
          - held_out_0.parquet:
        shape: (0, 2)
        ┌──────┬───────┐
        │ code ┆ count │
        │ ---  ┆ ---   │
        │ str  ┆ u32   │
        ╞══════╪═══════╡
        └──────┴───────┘
          - train_0.parquet:
        shape: (8, 2)
        ┌────────────────────┬───────┐
        │ code               ┆ count │
        │ ---                ┆ ---   │
        │ str                ┆ u32   │
        ╞════════════════════╪═══════╡
        │ EYE_COLOR//BROWN   ┆ 1     │
        │ HEIGHT             ┆ 2     │
        │ DOB                ┆ 2     │
        │ ADMISSION//CARDIAC ┆ 2     │
        │ HR                 ┆ 10    │
        │ TEMP               ┆ 10    │
        │ DISCHARGE          ┆ 2     │
        │ EYE_COLOR//BLUE    ┆ 1     │
        └────────────────────┴───────┘

    We can see that it keeps the files in this case (as it doesn't know they are only non-train files) but
    they are empty, as the non-train subjects have been removed. This is because the subject-split data is
    stored in the `metadata/subject_splits.parquet` file. If we remove this file as well, we'd get an error:

        >>> with tempfile.TemporaryDirectory() as MEDS_dir, tempfile.TemporaryDirectory() as out_dir:
        ...     copy_MEDS_without_split_sharding(Path(MEDS_dir))
        ...     (Path(MEDS_dir) / subject_splits_filepath).unlink()
        ...     cfg.stage_cfg.train_only = True
        ...     try:
        ...         profile_map_stage(out_dir, in_MEDS_dir=Path(MEDS_dir), map_fn=count_codes)
        ...     finally:
        ...         cfg.stage_cfg.train_only = False
        Traceback (most recent call last):
            ...
        FileNotFoundError: Train split requested, but shard prefixes can't be used and subject split file not
        found at /.../metadata/subject_splits.parquet.

    The examples so far have used the default shard iterator. You can read more about this in the
    `shard_iteration.py` file and its documentation, but we can also pass in our own shard iterators here, to
    show some more edge cases. For example, if we don't request train only, but the shard iterator asserts it
    is only returning train samples, that will cause an error:

        >>> def bad_shard_iterator(cfg: DictConfig) -> tuple[list[tuple[Path, Path]], bool]:
        ...     out_fps = []
        ...     for fp in Path(cfg.stage_cfg.data_input_dir).rglob("*.parquet"):
        ...         shard_name = fp.relative_to(cfg.stage_cfg.data_input_dir).as_posix()
        ...         if not shard_name.startswith("train/"):
        ...             # Skip the held-out and tuning splits as this is broken!
        ...             continue
        ...         out_fps.append((fp, cfg.stage_cfg.output_dir / shard_name))
        ...     return out_fps, True
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     cfg.stage_cfg.train_only = False
        ...     profile_map_stage(tmpdir, map_fn=count_codes, shard_iterator_fntr=bad_shard_iterator)
        Traceback (most recent call last):
            ...
        ValueError: All splits should be used, but shard iterator is returning only train splits?!?
    """

    start = datetime.now(tz=UTC)

    if read_fn is None:
        read_fn = read_df
    if write_fn is None:
        write_fn = write_df

    train_only = cfg.stage_cfg.get("train_only", False)

    shards, includes_only_train = shard_iterator_fntr(cfg)

    if train_only:
        split_fp = Path(cfg.input_dir) / subject_splits_filepath
        if includes_only_train:
            logger.info(
                f"Processing train split only via shard prefix. Not filtering with {split_fp.resolve()!s}."
            )
        elif split_fp.exists():
            logger.info(f"Processing train split only by filtering read dfs via {split_fp.resolve()!s}")
            train_subjects = (
                pl.scan_parquet(split_fp)
                .filter(pl.col(SubjectSplitSchema.split_name) == "train")
                .select(SubjectSplitSchema.subject_id_name)
                .collect()[SubjectSplitSchema.subject_id_name]
                .to_list()
            )
            read_fn = read_and_filter_fntr(pl.col(DataSchema.subject_id_name).is_in(train_subjects), read_fn)
        else:
            raise FileNotFoundError(
                f"Train split requested, but shard prefixes can't be used and "
                f"subject split file not found at {split_fp.resolve()!s}."
            )
    elif includes_only_train:
        raise ValueError("All splits should be used, but shard iterator is returning only train splits?!?")

    if is_match_revise(cfg.stage_cfg):
        map_fn = match_revise_fntr(cfg, cfg.stage_cfg, map_fn)
    else:
        map_fn = bind_compute_fn(cfg, cfg.stage_cfg, map_fn)

    all_out_fps = map_over(
        shards=shards,
        map_fn=map_fn,
        read_fn=read_fn,
        write_fn=write_fn,
        do_overwrite=cfg.do_overwrite,
    )
    logger.info(f"Finished mapping in {datetime.now(tz=UTC) - start}")
    return all_out_fps


def join_and_replace(new: DF_T, old: DF_T, join_cols: list[str]) -> DF_T:
    """Join two dataframes and replace the old columns with the new columns.

    Args:
        new: The new dataframe to join.
        old: The old dataframe to join.
        join_cols: The columns to join on.

    Returns:
        Adds the columns from `old` that are not in `new` into new, without replacing any values in `new` or
        changing the order of `new`, but while matching the join columns.

    Examples:
        >>> old = pl.DataFrame({"code": ["a", "b", "c"], "A": [1, 2, 3], "B": [4, 5, 6]})
        >>> new = pl.DataFrame({"code": ["c", "b", "d"], "A": [7, None, 9], "C": [10, 11, None]})
        >>> join_and_replace(new, old, ["code"])
        shape: (3, 4)
        ┌──────┬──────┬──────┬──────┐
        │ code ┆ A    ┆ C    ┆ B    │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ str  ┆ i64  ┆ i64  ┆ i64  │
        ╞══════╪══════╪══════╪══════╡
        │ c    ┆ 7    ┆ 10   ┆ 6    │
        │ b    ┆ null ┆ 11   ┆ 5    │
        │ d    ┆ 9    ┆ null ┆ null │
        └──────┴──────┴──────┴──────┘
    """

    new_cols = new.collect_schema().names()
    old_cols = old.collect_schema().names()

    return new.join(
        old.drop(*[c for c in old_cols if c in set(new_cols) - set(join_cols)]),
        on=join_cols,
        how="left",
        coalesce=True,
    )


def mapreduce_stage(
    cfg: DictConfig,
    map_fn: ANY_COMPUTE_FN_T,
    reduce_fn: REDUCE_FN_T,
    merge_fn: REDUCE_FN_T | None = None,
    read_fn: READ_FN_T | None = None,
    write_fn: WRITE_FN_T | None = None,
    shard_iterator_fntr: SHARD_ITR_FNTR_T = shard_iterator,
):
    """Performs a map-stage over shards produced by the shard iterator, then reduces over those outputs.

    Args:
        cfg: Configuration dictionary containing stage_cfg, input_dir, and other necessary parameters.
        map_fn: Function to apply to each shard.
        reduce_fn: Function to reduce the mapped data down to a single output.
        merge_fn: Function to merge the reduced data with the original data. Defaults to None, which resolves
            to `join_and_replace`, joining by the code column and any code modifiers.
        read_fn: Function to read data from the input file paths. Defaults to reading parquet files with
            polars.
        write_fn: Function to write the transformed data to the output file paths. Defaults to writing
            parquet files with polars.
        shard_iterator_fntr: Function to create the shard iterator. Defaults to the default shard iterator,
            which iterates over the parquet files in the data input directory within the `cfg.stage_cfg` in a
            pseudo-random, worker-dependent manner.

    Examples:
        >>> from meds_testing_helpers.dataset import MEDSDataset
        >>> D = MEDSDataset(root_dir=simple_static_MEDS)

    We'll show an example of this function using the `simple_static_MEDS` dataset provided as a pytest fixture
    and loaded in this doctest via our `conftest.py` file in the
    [`meds_testing_helpers`](https://github.com/Medical-Event-Data-Standard/meds_testing_helpers) package.

    To see this data, let's inspect it, first on disk (which is merely the typical MEDS fashion):

        >>> print_directory(simple_static_MEDS)
        ├── data
        │   ├── held_out
        │   │   └── 0.parquet
        │   ├── train
        │   │   ├── 0.parquet
        │   │   └── 1.parquet
        │   └── tuning
        │       └── 0.parquet
        └── metadata
            ├── codes.parquet
            ├── dataset.json
            └── subject_splits.parquet

    As well as its actual contents:

        >>> for k, df in D._pl_shards.items():
        ...     print(f"{k}:")
        ...     print(df)
        held_out/0:
        shape: (11, 4)
        ┌────────────┬─────────────────────┬───────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                  ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                   ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                   ┆ f32           │
        ╞════════════╪═════════════════════╪═══════════════════════╪═══════════════╡
        │ 1500733    ┆ null                ┆ EYE_COLOR//BROWN      ┆ null          │
        │ 1500733    ┆ null                ┆ HEIGHT                ┆ 158.601318    │
        │ 1500733    ┆ 1986-07-20 00:00:00 ┆ DOB                   ┆ null          │
        │ 1500733    ┆ 2010-06-03 14:54:38 ┆ ADMISSION//ORTHOPEDIC ┆ null          │
        │ 1500733    ┆ 2010-06-03 14:54:38 ┆ HR                    ┆ 91.400002     │
        │ …          ┆ …                   ┆ …                     ┆ …             │
        │ 1500733    ┆ 2010-06-03 15:39:49 ┆ HR                    ┆ 84.400002     │
        │ 1500733    ┆ 2010-06-03 15:39:49 ┆ TEMP                  ┆ 100.300003    │
        │ 1500733    ┆ 2010-06-03 16:20:49 ┆ HR                    ┆ 90.099998     │
        │ 1500733    ┆ 2010-06-03 16:20:49 ┆ TEMP                  ┆ 100.099998    │
        │ 1500733    ┆ 2010-06-03 16:44:26 ┆ DISCHARGE             ┆ null          │
        └────────────┴─────────────────────┴───────────────────────┴───────────────┘
        train/0:
        shape: (30, 4)
        ┌────────────┬─────────────────────┬────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code               ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                ┆ f32           │
        ╞════════════╪═════════════════════╪════════════════════╪═══════════════╡
        │ 239684     ┆ null                ┆ EYE_COLOR//BROWN   ┆ null          │
        │ 239684     ┆ null                ┆ HEIGHT             ┆ 175.271118    │
        │ 239684     ┆ 1980-12-28 00:00:00 ┆ DOB                ┆ null          │
        │ 239684     ┆ 2010-05-11 17:41:51 ┆ ADMISSION//CARDIAC ┆ null          │
        │ 239684     ┆ 2010-05-11 17:41:51 ┆ HR                 ┆ 102.599998    │
        │ …          ┆ …                   ┆ …                  ┆ …             │
        │ 1195293    ┆ 2010-06-20 20:24:44 ┆ HR                 ┆ 107.699997    │
        │ 1195293    ┆ 2010-06-20 20:24:44 ┆ TEMP               ┆ 100.0         │
        │ 1195293    ┆ 2010-06-20 20:41:33 ┆ HR                 ┆ 107.5         │
        │ 1195293    ┆ 2010-06-20 20:41:33 ┆ TEMP               ┆ 100.400002    │
        │ 1195293    ┆ 2010-06-20 20:50:04 ┆ DISCHARGE          ┆ null          │
        └────────────┴─────────────────────┴────────────────────┴───────────────┘
        train/1:
        shape: (14, 4)
        ┌────────────┬─────────────────────┬───────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                  ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                   ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                   ┆ f32           │
        ╞════════════╪═════════════════════╪═══════════════════════╪═══════════════╡
        │ 68729      ┆ null                ┆ EYE_COLOR//HAZEL      ┆ null          │
        │ 68729      ┆ null                ┆ HEIGHT                ┆ 160.395309    │
        │ 68729      ┆ 1978-03-09 00:00:00 ┆ DOB                   ┆ null          │
        │ 68729      ┆ 2010-05-26 02:30:56 ┆ ADMISSION//PULMONARY  ┆ null          │
        │ 68729      ┆ 2010-05-26 02:30:56 ┆ HR                    ┆ 86.0          │
        │ …          ┆ …                   ┆ …                     ┆ …             │
        │ 814703     ┆ 1976-03-28 00:00:00 ┆ DOB                   ┆ null          │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ ADMISSION//ORTHOPEDIC ┆ null          │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ HR                    ┆ 170.199997    │
        │ 814703     ┆ 2010-02-05 05:55:39 ┆ TEMP                  ┆ 100.099998    │
        │ 814703     ┆ 2010-02-05 07:02:30 ┆ DISCHARGE             ┆ null          │
        └────────────┴─────────────────────┴───────────────────────┴───────────────┘
        tuning/0:
        shape: (7, 4)
        ┌────────────┬─────────────────────┬──────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                 ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                  ┆ ---           │
        │ i64        ┆ datetime[μs]        ┆ str                  ┆ f32           │
        ╞════════════╪═════════════════════╪══════════════════════╪═══════════════╡
        │ 754281     ┆ null                ┆ EYE_COLOR//BROWN     ┆ null          │
        │ 754281     ┆ null                ┆ HEIGHT               ┆ 166.22261     │
        │ 754281     ┆ 1988-12-19 00:00:00 ┆ DOB                  ┆ null          │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ ADMISSION//PULMONARY ┆ null          │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ HR                   ┆ 142.0         │
        │ 754281     ┆ 2010-01-03 06:27:59 ┆ TEMP                 ┆ 99.800003     │
        │ 754281     ┆ 2010-01-03 08:22:13 ┆ DISCHARGE            ┆ null          │
        └────────────┴─────────────────────┴──────────────────────┴───────────────┘
        >>> D._pl_subject_splits
        shape: (6, 2)
        ┌────────────┬──────────┐
        │ subject_id ┆ split    │
        │ ---        ┆ ---      │
        │ i64        ┆ str      │
        ╞════════════╪══════════╡
        │ 239684     ┆ train    │
        │ 1195293    ┆ train    │
        │ 68729      ┆ train    │
        │ 814703     ┆ train    │
        │ 754281     ┆ tuning   │
        │ 1500733    ┆ held_out │
        └────────────┴──────────┘

    For mapreduce stages, the input code metadata is also relevant:

        >>> D._pl_code_metadata
        shape: (5, 3)
        ┌──────────────────┬─────────────────────────────────┬──────────────────┐
        │ code             ┆ description                     ┆ parent_codes     │
        │ ---              ┆ ---                             ┆ ---              │
        │ str              ┆ str                             ┆ list[str]        │
        ╞══════════════════╪═════════════════════════════════╪══════════════════╡
        │ EYE_COLOR//BLUE  ┆ Blue Eyes. Less common than br… ┆ null             │
        │ EYE_COLOR//BROWN ┆ Brown Eyes. The most common ey… ┆ null             │
        │ EYE_COLOR//HAZEL ┆ Hazel eyes. These are uncommon  ┆ null             │
        │ HR               ┆ Heart Rate                      ┆ ["LOINC/8867-4"] │
        │ TEMP             ┆ Body Temperature                ┆ ["LOINC/8310-5"] │
        └──────────────────┴─────────────────────────────────┴──────────────────┘

    We'll also make a simple helper too to print the output for us

        >>> def profile_mapreduce_stage(test_dir: str, worker: int | None = None, **kwargs):
        ...     '''Makes a test config, adds the output directory, runs the mapping stage, & shows outputs'''
        ...     test_cfg = DictConfig(cfg)
        ...     test_cfg.stage_cfg.output_dir = Path(test_dir) / "data_output"
        ...     test_cfg.stage_cfg.reducer_output_dir = Path(test_dir) / "reducer_output"
        ...     if worker is not None:
        ...         test_cfg.worker = worker
        ...     mapreduce_stage(cfg=test_cfg, **kwargs)
        ...     print("Data output directory:")
        ...     print_directory(test_cfg.stage_cfg.output_dir)
        ...     if test_cfg.stage_cfg.reducer_output_dir.exists():
        ...         print("Reducer output directory:")
        ...         print_directory(test_cfg.stage_cfg.reducer_output_dir)
        ...     out_fp = test_cfg.stage_cfg.reducer_output_dir / "codes.parquet"
        ...     if out_fp.exists():
        ...         print("------------------")
        ...         print("Reduced Output:")
        ...         print("------------------")
        ...         print(pl.read_parquet(out_fp))
        ...     else:
        ...         print("No reduced output found.")

    To mapreduce over this data, we need a configuration file that will point our data input dir to this
    dataset. Normally, the stage configuration is handled automatically by the Stage objects and the
    `PipelineConfig` class, but we'll just fudge it here for the sake of the example. Note we haven't added an
    output dir yet -- that will be a temporary directory we'll create just before we run the stage.

        >>> cfg = DictConfig(
        ...     {
        ...         "worker": 0,
        ...         "polling_time": 0.01,
        ...         "do_overwrite": False,
        ...         "input_dir": str(simple_static_MEDS),
        ...         "stage_cfg": {
        ...             "data_input_dir": str(simple_static_MEDS / "data"),
        ...             "metadata_input_dir": str(simple_static_MEDS / "metadata"),
        ...             "output_dir": "???",
        ...             "reducer_output_dir": "???",
        ...         },
        ...     }
        ... )

    We'll also need a mapping function that will be applied to each shard. For this example, we'll just count
    the number of occurrences of each code:

        >>> def count_codes(df: pl.LazyFrame) -> pl.LazyFrame:
        ...     return df.group_by("code", maintain_order=True).agg(pl.len().alias("count"))

    For our reducer function, we'll just sum the counts of each code across all shards:

        >>> def sum_counts(*dfs: pl.LazyFrame) -> pl.LazyFrame:
        ...     return pl.concat(dfs).group_by("code", maintain_order=True).agg(pl.sum("count"))

    Now we can run the mapreduce stage:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     profile_mapreduce_stage(tmpdir, map_fn=count_codes, reduce_fn=sum_counts)
        Data output directory:
        ├── held_out
        │   └── 0.parquet
        ├── train
        │   ├── 0.parquet
        │   └── 1.parquet
        └── tuning
            └── 0.parquet
        Reducer output directory:
        └── codes.parquet
        ------------------
        Reduced Output:
        ------------------
        shape: (11, 4)
        ┌───────────────────────┬───────┬─────────────────────────────────┬──────────────────┐
        │ code                  ┆ count ┆ description                     ┆ parent_codes     │
        │ ---                   ┆ ---   ┆ ---                             ┆ ---              │
        │ str                   ┆ u8    ┆ str                             ┆ list[str]        │
        ╞═══════════════════════╪═══════╪═════════════════════════════════╪══════════════════╡
        │ EYE_COLOR//BROWN      ┆ 3     ┆ Brown Eyes. The most common ey… ┆ null             │
        │ HEIGHT                ┆ 6     ┆ null                            ┆ null             │
        │ DOB                   ┆ 6     ┆ null                            ┆ null             │
        │ ADMISSION//ORTHOPEDIC ┆ 2     ┆ null                            ┆ null             │
        │ HR                    ┆ 16    ┆ Heart Rate                      ┆ ["LOINC/8867-4"] │
        │ …                     ┆ …     ┆ …                               ┆ …                │
        │ DISCHARGE             ┆ 6     ┆ null                            ┆ null             │
        │ ADMISSION//PULMONARY  ┆ 2     ┆ null                            ┆ null             │
        │ EYE_COLOR//HAZEL      ┆ 2     ┆ Hazel eyes. These are uncommon  ┆ null             │
        │ ADMISSION//CARDIAC    ┆ 2     ┆ null                            ┆ null             │
        │ EYE_COLOR//BLUE       ┆ 1     ┆ Blue Eyes. Less common than br… ┆ null             │
        └───────────────────────┴───────┴─────────────────────────────────┴──────────────────┘

    If we set `train_only` to `True`, we can see that the held-out and tuning splits are not included in the
    output. This is generally desired for stages aggregating over codes, as you don't want to include held out
    data in your normalization computation

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     cfg.stage_cfg.train_only = True
        ...     profile_mapreduce_stage(tmpdir, map_fn=count_codes, reduce_fn=sum_counts)
        ...     cfg.stage_cfg.train_only = False
        Data output directory:
        └── train
            ├── 0.parquet
            └── 1.parquet
        Reducer output directory:
        └── codes.parquet
        ------------------
        Reduced Output:
        ------------------
        shape: (11, 4)
        ┌───────────────────────┬───────┬─────────────────────────────────┬──────────────────┐
        │ code                  ┆ count ┆ description                     ┆ parent_codes     │
        │ ---                   ┆ ---   ┆ ---                             ┆ ---              │
        │ str                   ┆ u8    ┆ str                             ┆ list[str]        │
        ╞═══════════════════════╪═══════╪═════════════════════════════════╪══════════════════╡
        │ EYE_COLOR//HAZEL      ┆ 2     ┆ Hazel eyes. These are uncommon  ┆ null             │
        │ HEIGHT                ┆ 4     ┆ null                            ┆ null             │
        │ DOB                   ┆ 4     ┆ null                            ┆ null             │
        │ ADMISSION//PULMONARY  ┆ 1     ┆ null                            ┆ null             │
        │ HR                    ┆ 12    ┆ Heart Rate                      ┆ ["LOINC/8867-4"] │
        │ …                     ┆ …     ┆ …                               ┆ …                │
        │ DISCHARGE             ┆ 4     ┆ null                            ┆ null             │
        │ ADMISSION//ORTHOPEDIC ┆ 1     ┆ null                            ┆ null             │
        │ EYE_COLOR//BROWN      ┆ 1     ┆ Brown Eyes. The most common ey… ┆ null             │
        │ ADMISSION//CARDIAC    ┆ 2     ┆ null                            ┆ null             │
        │ EYE_COLOR//BLUE       ┆ 1     ┆ Blue Eyes. Less common than br… ┆ null             │
        └───────────────────────┴───────┴─────────────────────────────────┴──────────────────┘

    Lastly, note that mapreduce is special in that the reduction happens only in one worker -- so if the
    config has `worker` set to anything other than 0, it will exit after the mapping stage, and the reduction
    won't be completed:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     profile_mapreduce_stage(tmpdir, worker=1, map_fn=count_codes, reduce_fn=sum_counts)
        Data output directory:
        ├── held_out
        │   └── 0.parquet
        ├── train
        │   ├── 0.parquet
        │   └── 1.parquet
        └── tuning
            └── 0.parquet
        No reduced output found.
    """

    if read_fn is None:
        read_fn = read_df
    if write_fn is None:
        write_fn = write_df

    _out_fps = map_stage(
        cfg=cfg, map_fn=map_fn, read_fn=read_fn, write_fn=write_fn, shard_iterator_fntr=shard_iterator
    )

    if cfg.worker != 0:
        logger.info(f"Mapping completed. Exiting as am worker {cfg.worker}, not reducer (0).")
        return

    logger.info("Starting reduction process")
    start = datetime.now(tz=UTC)

    merge_fp = Path(cfg.stage_cfg.metadata_input_dir) / "codes.parquet"
    reduce_stage_out_fp = Path(cfg.stage_cfg.reducer_output_dir) / "codes.parquet"

    if merge_fn is None:
        join_cols = [DataSchema.code_name, *cfg.get("code_modifier_cols", [])]
        merge_fn = partial(join_and_replace, join_cols=join_cols)

    reduce_fn = bind_compute_fn(cfg, cfg.stage_cfg, reduce_fn)

    reduce_over(
        in_fps=_out_fps,
        out_fp=reduce_stage_out_fp,
        read_fn=read_fn,
        write_fn=write_fn,
        reduce_fn=reduce_fn,
        merge_fp=merge_fp,
        merge_fn=merge_fn,
        do_overwrite=cfg.do_overwrite,
        polling_time=cfg.polling_time,
    )
    logger.info(f"Finished reduction in {datetime.now(tz=UTC) - start}")



from collections.abc import Callable

import polars as pl
from meds import DataSchema
from omegaconf import DictConfig

from MEDS_transforms.compute_modes.compute_fn import identity_fn
from MEDS_transforms.stages import Stage


@Stage.register
def drop_regex(stage_cfg: DictConfig) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Drops the measurements whose codes match a given regex, or do nothing if no regex is provided.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 2, 3],
        ...     "time": [datetime(2020, 1, 1), datetime(2020, 1, 2), datetime(2020, 1, 3)],
        ...     "code": ["ICD10:A01", "ICD9:C03", "ICD9:D04"],
        ... })
        >>> drop_regex(DictConfig({"regex": "ICD10:.*"}))(df)
        shape: (2, 3)
        ┌────────────┬─────────────────────┬──────────┐
        │ subject_id ┆ time                ┆ code     │
        │ ---        ┆ ---                 ┆ ---      │
        │ i64        ┆ datetime[μs]        ┆ str      │
        ╞════════════╪═════════════════════╪══════════╡
        │ 2          ┆ 2020-01-02 00:00:00 ┆ ICD9:C03 │
        │ 3          ┆ 2020-01-03 00:00:00 ┆ ICD9:D04 │
        └────────────┴─────────────────────┴──────────┘
        >>> drop_regex(DictConfig({}))(df)
        shape: (3, 3)
        ┌────────────┬─────────────────────┬───────────┐
        │ subject_id ┆ time                ┆ code      │
        │ ---        ┆ ---                 ┆ ---       │
        │ i64        ┆ datetime[μs]        ┆ str       │
        ╞════════════╪═════════════════════╪═══════════╡
        │ 1          ┆ 2020-01-01 00:00:00 ┆ ICD10:A01 │
        │ 2          ┆ 2020-01-02 00:00:00 ┆ ICD9:C03  │
        │ 3          ┆ 2020-01-03 00:00:00 ┆ ICD9:D04  │
        └────────────┴─────────────────────┴───────────┘
    """

    regex = stage_cfg.get("regex", None)

    if not regex:
        return identity_fn

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        return df.filter(~pl.col(DataSchema.code_name).str.contains(regex))

    return fn



"""This script is a helper utility to run entire pipelines from a single script.

To do this effectively, this runner functionally takes a "meta configuration" file that contains:

  1. The path to the pipeline configuration file.
  2. Configuration details for how to run each stage of the pipeline, including mappings to the underlying
     stage scripts and Hydra launcher configurations for each stage to control parallelism, resources, etc.
"""

import argparse
import logging
import subprocess
from pathlib import Path

import yaml
from omegaconf import DictConfig, OmegaConf

from .configs import PipelineConfig
from .configs.utils import OmegaConfResolver

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

logger = logging.getLogger(__name__)


def get_parallelization_args(
    parallelization_cfg: dict | DictConfig | None, default_parallelization_cfg: dict | DictConfig
) -> list[str]:
    """Extracts the specific parallelization arguments given the default and stage-specific configurations.

    Args:
        parallelization_cfg: The stage-specific parallelization configuration.
        default_parallelization_cfg: The default parallelization configuration.

    Returns:
        A list of command-line arguments for parallelization.

    Examples:
        >>> get_parallelization_args({}, {})
        []
        >>> get_parallelization_args(None, {"n_workers": 4})
        []
        >>> get_parallelization_args({"launcher": "joblib"}, {})
        ['--multirun', 'worker="range(0,1)"', 'hydra/launcher=joblib']
        >>> get_parallelization_args({"n_workers": 2, "launcher_params": "foo"}, {})
        Traceback (most recent call last):
            ...
        ValueError: If launcher_params is provided, launcher must also be provided.
        >>> get_parallelization_args({"n_workers": 2}, {})
        ['--multirun', 'worker="range(0,2)"']
        >>> get_parallelization_args({"launcher": "slurm"}, {"n_workers": 3, "launcher": "joblib"})
        ['--multirun', 'worker="range(0,3)"', 'hydra/launcher=slurm']
        >>> get_parallelization_args(
        ...     {"n_workers": 2, "launcher": "joblib"},
        ...     {"n_workers": 5, "launcher_params": {"foo": "bar"}},
        ... )
        ['--multirun', 'worker="range(0,2)"', 'hydra/launcher=joblib', 'hydra.launcher.foo=bar']
        >>> get_parallelization_args(
        ...     {"n_workers": 5, "launcher_params": {"biz": "baz"}, "launcher": "slurm"}, {}
        ... )
        ['--multirun', 'worker="range(0,5)"', 'hydra/launcher=slurm', 'hydra.launcher.biz=baz']
    """

    if parallelization_cfg is None:
        return []

    if len(parallelization_cfg) == 0 and len(default_parallelization_cfg) == 0:
        return []

    if "n_workers" in parallelization_cfg:
        n_workers = parallelization_cfg["n_workers"]
    elif "n_workers" in default_parallelization_cfg:
        n_workers = default_parallelization_cfg["n_workers"]
    else:
        n_workers = 1

    parallelization_args = [
        "--multirun",
        f'worker="range(0,{n_workers})"',
    ]

    if "launcher" in parallelization_cfg:
        launcher = parallelization_cfg["launcher"]
    elif "launcher" in default_parallelization_cfg:
        launcher = default_parallelization_cfg["launcher"]
    else:
        launcher = None

    if launcher is None:
        if "launcher_params" in parallelization_cfg:
            raise ValueError("If launcher_params is provided, launcher must also be provided.")

        return parallelization_args

    parallelization_args.append(f"hydra/launcher={launcher}")

    if "launcher_params" in parallelization_cfg:
        launcher_params = parallelization_cfg["launcher_params"]
    elif "launcher_params" in default_parallelization_cfg:
        launcher_params = default_parallelization_cfg["launcher_params"]
    else:
        launcher_params = {}

    for k, v in launcher_params.items():
        parallelization_args.append(f"hydra.launcher.{k}={v}")

    return parallelization_args


def run_stage(
    pipeline_config_fp: str,
    stage_runners_cfg: dict | DictConfig,
    pipeline_cfg: PipelineConfig,
    stage_name: str,
    cfg_overrides: list[str] | None = None,
    default_parallelization_cfg: dict | DictConfig | None = None,
    do_profile: bool = False,
    runner_fn: callable = subprocess.run,  # For dependency injection
):
    """Runs a single stage of the pipeline.

    Args:
        pipeline_config_fp: Path to the pipeline configuration on disk.
        stage_runners_cfg: The dictionary of stage runner configurations.
        stage_name: The name of the stage to run.
        default_parallelization_cfg: Default parallelization configuration.
        do_profile: Whether to enable Hydra profiling for the stage.
        runner_fn: Function used to actually run the stage command. This is
            primarily for dependency injection in testing.

    Raises:
        ValueError: If the stage fails to run.

    Examples:
        >>> def fake_shell_succeed(cmd, shell, capture_output):
        ...     print(cmd)
        ...     return subprocess.CompletedProcess(args=cmd, returncode=0, stdout=b"", stderr=b"")
        >>> def fake_shell_fail(cmd, shell, capture_output):
        ...     print(cmd)
        ...     return subprocess.CompletedProcess(args=cmd, returncode=1, stdout=b"", stderr=b"")
        >>> stage_runners = {
        ...     "reshard_to_split": {"_script": "not used"},
        ...     "fit_vocabulary_indices": {},
        ...     "reorder_measurements": {"script": "baz_script"},
        ... }
        >>> pipeline_cfg = PipelineConfig(
        ...     stages=[
        ...         "reshard_to_split",
        ...         {"fit_vocabulary_indices": {"_script": "foobar"}},
        ...         "reorder_measurements",
        ...     ],
        ... )
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "reshard_to_split",
        ...     runner_fn=fake_shell_succeed,
        ... )
        MEDS_transform-stage pipeline_config.yaml reshard_to_split stage=reshard_to_split
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "fit_vocabulary_indices",
        ...     runner_fn=fake_shell_succeed,
        ... )
        foobar stage=fit_vocabulary_indices
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "reorder_measurements",
        ...     runner_fn=fake_shell_succeed,
        ... )
        baz_script stage=reorder_measurements
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "reorder_measurements",
        ...     do_profile=True,
        ...     runner_fn=fake_shell_succeed,
        ... )
        baz_script stage=reorder_measurements
            ++hydra.callbacks.profiler._target_=hydra_profiler.profiler.ProfilerCallback
        >>> stage_runners["reorder_measurements"]["parallelize"] = {"n_workers": 2}
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "reorder_measurements",
        ...     runner_fn=fake_shell_succeed,
        ... )
        baz_script --multirun stage=reorder_measurements worker="range(0,2)"
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     stage_runners,
        ...     pipeline_cfg,
        ...     "reorder_measurements",
        ...     runner_fn=fake_shell_fail,
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Stage reorder_measurements failed via ...

        Improper configurations also raise errors:

        >>> bad_runners = {"reshard_to_split": {"_base_stage": "belongs in the stage"}}
        >>> bad_cfg = PipelineConfig(stages=["reshard_to_split"])
        >>> run_stage(
        ...     "pipeline_config.yaml",
        ...     bad_runners,
        ...     bad_cfg,
        ...     "reshard_to_split",
        ...     runner_fn=fake_shell_succeed,
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Put _base_stage args is in your pipeline config
    """

    if default_parallelization_cfg is None:
        default_parallelization_cfg = {}

    stage_config = pipeline_cfg.parsed_stages_by_name[stage_name].config
    stage_runner_config = stage_runners_cfg.get(stage_name, {})

    script = None
    if "script" in stage_runner_config:
        script = stage_runner_config.get("script")
    elif "_script" in stage_config:
        script = stage_config.get("_script")
    elif "_base_stage" in stage_runner_config:
        raise ValueError("Put _base_stage args is in your pipeline config")
    else:
        script = f"MEDS_transform-stage {pipeline_config_fp} {stage_name}"

    command_parts = [
        script,
        f"stage={stage_name}",
    ]

    if cfg_overrides:
        command_parts.extend(cfg_overrides)

    parallelization_args = get_parallelization_args(
        stage_runner_config.get("parallelize", {}), default_parallelization_cfg
    )

    if parallelization_args:
        multirun = parallelization_args.pop(0)
        command_parts = command_parts[:1] + [multirun] + command_parts[1:] + parallelization_args

    if do_profile:
        command_parts.append("++hydra.callbacks.profiler._target_=hydra_profiler.profiler.ProfilerCallback")

    full_cmd = " ".join(command_parts)
    logger.info(f"Running command: {full_cmd}")
    command_out = runner_fn(full_cmd, shell=True, capture_output=True)

    # https://stackoverflow.com/questions/21953835/run-subprocess-and-print-output-to-logging

    stderr = command_out.stderr.decode()
    stdout = command_out.stdout.decode()
    logger.info(f"Command output:\n{stdout}")
    logger.info(f"Command error:\n{stderr}")

    if command_out.returncode != 0:
        raise ValueError(
            f"Stage {stage_name} failed via {full_cmd} with return code {command_out.returncode}."
        )


def main(argv: list[str] | None = None) -> int:
    """Run an entire pipeline based on command line arguments."""

    parser = argparse.ArgumentParser(description="MEDS-Transforms Pipeline Runner")
    parser.add_argument(
        "pipeline_config_fp",
        help="Path to the pipeline configuration file, either as a raw path or with pkg:// syntax.",
    )
    parser.add_argument(
        "--stage_runner_fp",
        type=str,
        default=None,
        help="Path to the stage runner configuration file. If not provided, no specialized runner is used.",
    )
    parser.add_argument(
        "--do_profile",
        default=False,
        type=bool,
        action=argparse.BooleanOptionalAction,
        help="Enable Hydra profiling for the stages.",
    )
    parser.add_argument(
        "--overrides", nargs="*", default=[], help="Additional overrides for the pipeline configuration."
    )
    args = parser.parse_args(argv)

    pipeline_config = PipelineConfig.from_arg(args.pipeline_config_fp, args.overrides)
    if pipeline_config.additional_params is None or "output_dir" not in pipeline_config.additional_params:
        raise ValueError("Pipeline configuration or override must specify an 'output_dir'")

    log_dir = Path(pipeline_config.additional_params["output_dir"]) / ".logs"
    log_dir.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(filename=(log_dir / "pipeline.log"), level=logging.INFO)

    logging.info("Running MEDS-Transforms Pipeline Runner with the following arguments:")
    for arg_name, arg_value in vars(args).items():
        logging.info(f"  {arg_name}: {arg_value}")

    global_done_file = log_dir / "_all_stages.done"
    if global_done_file.exists():
        logger.info("All stages are already complete. Exiting.")
        return 0

    stages = [s.name for s in pipeline_config.parsed_stages]
    if not stages:
        raise ValueError("Pipeline configuration must specify at least one stage.")

    stage_runners_cfg = load_yaml_file(args.stage_runner_fp)

    if "parallelize" in stage_runners_cfg:
        logger.info("Parallelization configuration loaded from stage runner")
        default_parallelization_cfg = stage_runners_cfg["parallelize"]
    elif "parallelize" in pipeline_config.additional_params:
        logger.info("Parallelization configuration loaded from pipeline config")
        default_parallelization_cfg = pipeline_config.additional_params["parallelize"]
    else:
        logger.info("No parallelization configuration provided.")
        default_parallelization_cfg = None

    for stage in stages:
        done_file = log_dir / f"{stage}.done"

        if done_file.exists():
            logger.info(f"Skipping stage {stage} as it is already complete.")
        else:
            logger.info(f"Running stage: {stage}")
            run_stage(
                args.pipeline_config_fp,
                stage_runners_cfg,
                pipeline_config,
                stage,
                cfg_overrides=args.overrides,
                default_parallelization_cfg=default_parallelization_cfg,
                do_profile=args.do_profile,
            )
            done_file.touch()

    global_done_file.touch()
    return 0


@OmegaConfResolver
def load_yaml_file(path: str | None) -> dict | DictConfig:
    """Loads a YAML file as an OmegaConf object.

    Args:
        path: The path to the YAML file.

    Returns:
        The OmegaConf object representing the YAML file, or None if no path is provided.

    Raises:
        FileNotFoundError: If the file does not exist.

    Examples:
        >>> load_yaml_file(None)
        {}
        >>> load_yaml_file("nonexistent_file.yaml")
        Traceback (most recent call last):
            ...
        FileNotFoundError: File nonexistent_file.yaml does not exist.
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
        ...     _ = f.write(b"foo: bar")
        ...     f.flush()
        ...     load_yaml_file(f.name)
        {'foo': 'bar'}
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
        ...     cfg = OmegaConf.create({"foo": "bar"})
        ...     OmegaConf.save(cfg, f.name)
        ...     load_yaml_file(f.name)
        {'foo': 'bar'}
    """

    if not path:
        return {}

    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"File {path} does not exist.")

    try:
        return OmegaConf.load(path)
    except Exception as e:  # pragma: no cover
        logger.warning(f"Failed to load {path} as an OmegaConf: {e}. Trying as a plain YAML file.")
        yaml_text = path.read_text()
        return yaml.load(yaml_text, Loader=Loader)



"""Utility helpers for path resolution and miscellaneous tasks."""

import os
from importlib.resources import files
from pathlib import Path

PKG_PFX = "pkg://"


def resolve_pkg_path(pkg_path: str) -> Path:
    """Resolve a ``pkg://`` path into an on-disk :class:`~pathlib.Path`.

    Args:
        pkg_path: Path in ``pkg://`` notation.

    Returns:
        The resolved path to the package resource on disk.

    Raises:
        ValueError: If the package specified in ``pkg_path`` does not exist.

    Examples:
        >>> resolve_pkg_path("pkg://MEDS_transforms.configs.pipeline.py").suffix
        '.py'
    """
    parts = pkg_path[len(PKG_PFX) :].split(".")
    pkg_name = parts[0]
    suffix = parts[-1]
    relative_path = Path(os.path.join(*parts[1:-1])).with_suffix(f".{suffix}")
    try:
        return files(pkg_name) / relative_path
    except ModuleNotFoundError as e:
        raise ValueError(f"Package '{pkg_name}' not found. Please check the package name.") from e



from importlib.metadata import version

__package_name__ = __package__
__version__ = version(__package_name__)


INFERRED_STAGE_KEYS = {
    "is_metadata",
    "train_only",
    "data_input_dir",
    "metadata_input_dir",
    "output_dir",
    "reducer_output_dir",
    "_script",
}



"""A functor to compute timeline token measurements on a MEDS dataset."""

import logging
from collections.abc import Callable

import polars as pl
from meds import DataSchema
from omegaconf import DictConfig

from ...compute_modes.compute_fn import identity_fn
from .utils import normalize_time_unit, unique_events

logger = logging.getLogger(__name__)


def timeline_tokens_fntr(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
    """Create a function that adds timeline-measurement rows between events (distinct timepoints).

    In this context, "timeline-measurements" correspond to added measurements that capture properties about
    the timeline of the subject's events -- namely, the start of their longitudinal (non-static) data, the
    time between subsequent events in their data, and the end of their longitudinal data.

    Args:
        cfg: The configuration for the time delta function. This must contain the following keys:
            - "time_delta_code": The code for time-delta measurements in the output data. If not provided,
              defaults to `"TIMELINE//DELTA//{time_unit}"`. If the code string provided contains
              `{time_unit}`, it will be substituted for the specified time unit (in normalized form). If
              `null`, these rows will not be added.
            - "time_unit": The unit for the time deltas when dates are converted to a numeric value in the
              output data. This is mandatory of time delta codes are added at all. For the full list of units
              supported, see `utils.normalize_time_unit`.
            - "timeline_start_code": The code for the "start of temporal data" code in the output data. If
              `null`, these rows will not be added. If not provided, defaults to `"TIMELINE//START"`.
            - "timeline_end_code": The code for the "end of temporal data" code in the output data. If `null`,
              these rows will not be added. If not provided, defaults to `"TIMELINE//END"`.

    > [!NOTE]
    > If a subject has only static data, no time delta or timeline start/end tokens will be added.

    Returns:
        A function that returns the to-be-added "time_delta" measurements between subsequent events in a
        subject's data. The very first event for a subject has a null time_delta, so it is imputed with a
        special time start token.

    Raises:
        ValueError: If the input unit is not recognized.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 1, 2, 2, 3, 3],
        ...         "time": [
        ...             None,
        ...             datetime(1990, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 2),
        ...             datetime(1988, 1, 2),
        ...             datetime(2023, 1, 3),
        ...             None,
        ...             None,
        ...         ],
        ...         "code": ["static", "DOB", "lab//A", "lab//B", "rx", "DOB", "lab//A", "foo", "bar"],
        ...     },
        ...     schema={"subject_id": pl.UInt32, "time": pl.Datetime, "code": pl.Utf8},
        ... )
        >>> df
        shape: (9, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ u32        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB    │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//B │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ rx     │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ DOB    │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ lab//A │
        │ 3          ┆ null                ┆ foo    │
        │ 3          ┆ null                ┆ bar    │
        └────────────┴─────────────────────┴────────┘

    If we simply provide a time unit, we get start, end, and delta tokens in the specified unit. Note the
    output column name depends on the normalized time unit.

        >>> timeline_tokens_fn = timeline_tokens_fntr(DictConfig({"time_unit": "day"}))
        >>> timeline_tokens_fn(df)
        shape: (7, 4)
        ┌────────────┬─────────────────────┬───────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                  ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                   ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str                   ┆ f32           │
        ╞════════════╪═════════════════════╪═══════════════════════╪═══════════════╡
        │ 1          ┆ 1990-01-01 00:00:00 ┆ TIMELINE//START       ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ TIMELINE//DELTA//days ┆ 11323.0       │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ TIMELINE//DELTA//days ┆ 1.0           │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ TIMELINE//END         ┆ null          │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ TIMELINE//START       ┆ null          │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//DELTA//days ┆ 12785.0       │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//END         ┆ null          │
        └────────────┴─────────────────────┴───────────────────────┴───────────────┘

    We can change the codes for any of the timeline token types and the unit used for the time deltas via the
    config:

        >>> cfg = DictConfig({
        ...     "time_unit": "y",
        ...     "timeline_start_code": "ST",
        ...     "timeline_end_code": "END",
        ...     "time_delta_code": "DELTA//{time_unit}",
        ... })
        >>> timeline_tokens_fn = timeline_tokens_fntr(cfg)
        >>> timeline_tokens_fn(df)
        shape: (7, 4)
        ┌────────────┬─────────────────────┬──────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code         ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---          ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str          ┆ f32           │
        ╞════════════╪═════════════════════╪══════════════╪═══════════════╡
        │ 1          ┆ 1990-01-01 00:00:00 ┆ ST           ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ DELTA//years ┆ 31.001347     │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ DELTA//years ┆ 0.002738      │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ END          ┆ null          │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ ST           ┆ null          │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ DELTA//years ┆ 35.004169     │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ END          ┆ null          │
        └────────────┴─────────────────────┴──────────────┴───────────────┘

    We can also set the time code for deltas to not contain the unit string:

        >>> cfg = DictConfig({
        ...     "time_unit": "mos", # Months
        ...     "time_delta_code": "DELTA",
        ... })
        >>> timeline_tokens_fn = timeline_tokens_fntr(cfg)
        >>> timeline_tokens_fn(df)
        shape: (7, 4)
        ┌────────────┬─────────────────────┬─────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code            ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---             ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str             ┆ f32           │
        ╞════════════╪═════════════════════╪═════════════════╪═══════════════╡
        │ 1          ┆ 1990-01-01 00:00:00 ┆ TIMELINE//START ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ DELTA           ┆ 372.015839    │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ DELTA           ┆ 0.032855      │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ TIMELINE//END   ┆ null          │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ TIMELINE//START ┆ null          │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ DELTA           ┆ 420.049683    │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//END   ┆ null          │
        └────────────┴─────────────────────┴─────────────────┴───────────────┘

    We can disable any or all of the components by setting their code field to `null`.

        >>> timeline_tokens_fn = timeline_tokens_fntr(DictConfig({"time_delta_code": None}))
        >>> timeline_tokens_fn(df)
        shape: (4, 3)
        ┌────────────┬─────────────────────┬─────────────────┐
        │ subject_id ┆ time                ┆ code            │
        │ ---        ┆ ---                 ┆ ---             │
        │ u32        ┆ datetime[μs]        ┆ str             │
        ╞════════════╪═════════════════════╪═════════════════╡
        │ 1          ┆ 1990-01-01 00:00:00 ┆ TIMELINE//START │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ TIMELINE//END   │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ TIMELINE//START │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//END   │
        └────────────┴─────────────────────┴─────────────────┘
        >>> cfg = DictConfig({
        ...     "time_unit": "year",
        ...     "timeline_start_code": None,
        ...     "timeline_end_code": None,
        ... })
        >>> timeline_tokens_fn = timeline_tokens_fntr(cfg)
        >>> timeline_tokens_fn(df)
        shape: (3, 4)
        ┌────────────┬─────────────────────┬────────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                   ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                    ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str                    ┆ f32           │
        ╞════════════╪═════════════════════╪════════════════════════╪═══════════════╡
        │ 1          ┆ 2021-01-01 00:00:00 ┆ TIMELINE//DELTA//years ┆ 31.001347     │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ TIMELINE//DELTA//years ┆ 0.002738      │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//DELTA//years ┆ 35.004169     │
        └────────────┴─────────────────────┴────────────────────────┴───────────────┘

    If all are None, the function will return an identity function that does not modify the input
    argument.

        >>> cfg = DictConfig({
        ...     "timeline_start_code": None,
        ...     "timeline_end_code": None,
        ...     "time_delta_code": None,
        ... })
        >>> timeline_tokens_fn = timeline_tokens_fntr(cfg)
        >>> timeline_tokens_fn(df)
        shape: (9, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ u32        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB    │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//B │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ rx     │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ DOB    │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ lab//A │
        │ 3          ┆ null                ┆ foo    │
        │ 3          ┆ null                ┆ bar    │
        └────────────┴─────────────────────┴────────┘
        >>> timeline_tokens_fn("foobar") # This won't raise an error as this is truly an identity function.
        'foobar'
    """

    timeline_start_token = cfg.get("timeline_start_code", "TIMELINE//START")
    do_add_timeline_start = timeline_start_token is not None

    timeline_end_token = cfg.get("timeline_end_code", "TIMELINE//END")
    do_add_timeline_end = timeline_end_token is not None

    time_delta_token = cfg.get("time_delta_code", "TIMELINE//DELTA//{time_unit}")
    do_add_time_delta = time_delta_token is not None

    if do_add_time_delta:
        normalized_unit, seconds_in_unit = normalize_time_unit(cfg.time_unit)
        microseconds_in_unit = int(1e6) * seconds_in_unit
        time_delta_token = time_delta_token.format(time_unit=normalized_unit)

    if not (do_add_timeline_start or do_add_time_delta or do_add_timeline_end):
        logger.warning(
            "No timeline tokens were configured to be added. This is likely an issue with your config. "
            "Returning an identity function."
        )
        return identity_fn

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        events = unique_events(df)

        out_measurements = []

        time = pl.col(DataSchema.time_name)
        code_dtype = df.collect_schema().get(DataSchema.code_name, pl.Utf8)
        numeric_value_dtype = df.collect_schema().get(DataSchema.numeric_value_name, pl.Float32)

        if do_add_timeline_start:
            first_time = time.min().over(DataSchema.subject_id_name)
            timeline_start_code = pl.lit(timeline_start_token, dtype=code_dtype)

            out_measurements.append(
                events.filter(pl.col(DataSchema.time_name) == first_time).with_columns(
                    timeline_start_code.alias(DataSchema.code_name)
                )
            )

        if do_add_time_delta:
            time_delta_code = pl.lit(time_delta_token, dtype=code_dtype)

            time_delta = time.diff(null_behavior="ignore").over(DataSchema.subject_id_name)
            timeline_deltas_expr = (time_delta.dt.total_microseconds() / microseconds_in_unit).cast(
                numeric_value_dtype
            )

            out_measurements.append(
                events.with_columns(
                    time_delta_code.alias(DataSchema.code_name),
                    timeline_deltas_expr.alias(DataSchema.numeric_value_name),
                ).filter(pl.col(DataSchema.numeric_value_name).is_not_null())
            )

        if do_add_timeline_end:
            last_time = time.max().over(DataSchema.subject_id_name)
            timeline_end_code = pl.lit(timeline_end_token, dtype=code_dtype)

            out_measurements.append(
                events.filter(pl.col(DataSchema.time_name) == last_time).with_columns(
                    timeline_end_code.alias(DataSchema.code_name)
                )
            )

        return pl.concat(out_measurements, how="diagonal").sort(
            by=[DataSchema.subject_id_name, DataSchema.time_name], maintain_order=True
        )

    return fn



from .base import Stage
from .discovery import get_all_registered_stages
from .examples import StageExample

# Here are all the stages that are registered in the entry points, imported here so they can be imported at a
# module level.
# isort: split
from .add_time_derived_measurements import stage as add_time_derived_measurements
from .aggregate_code_metadata import stage as aggregate_code_metadata
from .bin_numeric_values import stage as bin_numeric_values
from .docgen import StageDoc, generate_stage_docs
from .extract_values import stage as extract_values
from .filter_measurements import stage as filter_measurements
from .filter_subjects import stage as filter_subjects
from .fit_vocabulary_indices import stage as fit_vocabulary_indices
from .normalization import stage as normalization
from .occlude_outliers import stage as occlude_outliers
from .reorder_measurements import stage as reorder_measurements
from .reshard_to_split import stage as reshard_to_split
